

from pydantic import BaseModel, validator, ValidationError
from contentctl.objects.mitre_attack_enrichment import MitreAttackEnrichment


class StoryTags(BaseModel):
    # story spec
    name: str
    analytic_story: str
    category: list
    product: list
    usecase: str

    # enrichment
    mitre_attack_enrichments: list[MitreAttackEnrichment] = []
    mitre_attack_tactics: list = []
    datamodels: list = []
    kill_chain_phases: list = []


    @validator('product')
    def tags_product(cls, v, values):
        valid_products = [
            "Splunk Enterprise", "Splunk Enterprise Security", "Splunk Cloud",
            "Splunk Security Analytics for AWS", "Splunk Behavioral Analytics"
        ]

        for value in v:
            if value not in valid_products:
                raise ValueError('product is not valid for ' + values['name'] + '. valid products are ' + str(valid_products))
        return v



import json


VERSION = "4.3"
NAME = "Detection Coverage"
DESCRIPTION = "security_content detection coverage"
DOMAIN = "mitre-enterprise"


class AttackNavWriter():

    @staticmethod
    def writeAttackNavFile(mitre_techniques : dict, output_path : str) -> None:
        max_count = 0
        for technique_id in mitre_techniques.keys():
            if mitre_techniques[technique_id]['score'] > max_count:
                max_count = mitre_techniques[technique_id]['score']
        
        layer_json = {
            "version": VERSION,
            "name": NAME,
            "description": DESCRIPTION,
            "domain": DOMAIN,
            "techniques": []
        }

        layer_json["gradient"] = {
            "colors": [
                "#ffffff",
                "#66b1ff",
                "#096ed7"
            ],
            "minValue": 0,
            "maxValue": max_count
        }

        layer_json["filters"] = {
                "platforms":
                    ["Windows",
                    "Linux",
                    "macOS",
                    "AWS",
                    "GCP",
                    "Azure",
                    "Office 365",
                    "SaaS"
                ]
        }

        layer_json["legendItems"] = [
            {
                "label": "NO available detections",
                "color": "#ffffff"
            },
            {
                "label": "Some detections available",
                "color": "#66b1ff"
            }
        ]

        layer_json['showTacticRowBackground'] = True
        layer_json['tacticRowBackground'] = "#dddddd"
        layer_json["sorting"] = 3

        for technique_id in mitre_techniques.keys():
            layer_technique = {
                "techniqueID": technique_id,
                "score": mitre_techniques[technique_id]['score'],
                "comment": "\n\n".join(mitre_techniques[technique_id]['file_paths'])
            }
            layer_json["techniques"].append(layer_technique)

        with open(output_path, 'w') as outfile:
            json.dump(layer_json, outfile, ensure_ascii=False, indent=4)



from dataclasses import dataclass
import os
import glob
import shutil
import tarfile
from typing import Union
from pathlib import Path

from contentctl.output.conf_writer import ConfWriter
from contentctl.objects.enums import SecurityContentType
from contentctl.objects.config import Config


class ConfOutput:
    input_path: str
    output_path: str
    app_name: str

    def __init__(self, input_path: str, config: Config):
        self.input_path = input_path
        self.app_name = config.build.splunk_app.prefix
        self.output_path = os.path.join(input_path, config.build.splunk_app.path)
        Path(self.output_path).mkdir(parents=True, exist_ok=True)
        template_splunk_app_path = os.path.join(os.path.dirname(__file__), '../templates/splunk_app')
        shutil.copytree(template_splunk_app_path, self.output_path, dirs_exist_ok=True)


    def writeHeaders(self) -> None:
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/analyticstories.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/savedsearches.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/collections.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/es_investigations.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/macros.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/transforms.conf'))
        ConfWriter.writeConfFileHeader(os.path.join(self.output_path, 'default/workflow_actions.conf'))


    def writeObjects(self, objects: list, type: SecurityContentType = None) -> None:
        if type == SecurityContentType.detections:
            ConfWriter.writeConfFile('savedsearches_detections.j2', 
            os.path.join(self.output_path, 'default/savedsearches.conf'), 
            objects,self.app_name)

            ConfWriter.writeConfFile('analyticstories_detections.j2',
                os.path.join(self.output_path, 'default/analyticstories.conf'), 
                objects,self.app_name)

            ConfWriter.writeConfFile('macros_detections.j2',
                os.path.join(self.output_path, 'default/macros.conf'), 
                objects,self.app_name)
        
        elif type == SecurityContentType.stories:
            ConfWriter.writeConfFile('analyticstories_stories.j2',
                os.path.join(self.output_path, 'default/analyticstories.conf'), 
                objects,self.app_name)

        elif type == SecurityContentType.baselines:
            ConfWriter.writeConfFile('savedsearches_baselines.j2', 
                os.path.join(self.output_path, 'default/savedsearches.conf'), 
                objects,self.app_name)

        elif type == SecurityContentType.investigations:
            ConfWriter.writeConfFile('savedsearches_investigations.j2', 
                os.path.join(self.output_path, 'default/savedsearches.conf'), 
                objects,self.app_name)
            
            ConfWriter.writeConfFile('analyticstories_investigations.j2', 
                os.path.join(self.output_path, 'default/analyticstories.conf'), 
                objects,self.app_name)

            workbench_panels = []
            for investigation in objects:
                if investigation.inputs:
                    response_file_name_xml = investigation.lowercase_name + "___response_task.xml"
                    workbench_panels.append(investigation)
                    investigation.search = investigation.search.replace(">","&gt;")
                    investigation.search = investigation.search.replace("<","&lt;")
                    ConfWriter.writeConfFileHeaderEmpty(os.path.join(self.output_path, 
                        'default/data/ui/panels/', str("workbench_panel_" + response_file_name_xml)))
                    ConfWriter.writeConfFile('panel.j2', 
                        os.path.join(self.output_path, 
                        'default/data/ui/panels/', str("workbench_panel_" + response_file_name_xml)),
                        [investigation.search],self.app_name)

            ConfWriter.writeConfFile('es_investigations_investigations.j2', 
                os.path.join(self.output_path, 'default/es_investigations.conf'), 
                workbench_panels,self.app_name)

            ConfWriter.writeConfFile('workflow_actions.j2', 
                os.path.join(self.output_path, 'default/workflow_actions.conf'), 
                workbench_panels,self.app_name)   

        elif type == SecurityContentType.lookups:
            ConfWriter.writeConfFile('collections.j2', 
                os.path.join(self.output_path, 'default/collections.conf'), 
                objects,self.app_name)

            ConfWriter.writeConfFile('transforms.j2', 
                os.path.join(self.output_path, 'default/transforms.conf'), 
                objects,self.app_name)


            if self.input_path is None:
                raise(Exception(f"input_path is required for lookups, but received [{self.input_path}]"))

            files = glob.iglob(os.path.join(self.input_path, 'lookups', '*.csv'))
            for file in files:
                if os.path.isfile(file):
                    shutil.copy(file, os.path.join(self.output_path, 'lookups'))

        elif type == SecurityContentType.macros:
            ConfWriter.writeConfFile('macros.j2', 
                os.path.join(self.output_path, 'default/macros.conf'), 
                objects,self.app_name)


    def packageApp(self) -> None:
        name = self.output_path + ".tar.gz"

        with tarfile.open(name, "w:gz") as app_archive:
            app_archive.add(self.output_path, arcname=os.path.basename(self.output_path))



import argparse
import json
import sys

RAW_BADGE_SVG = '''<?xml version="1.0"?>
<svg xmlns="http://www.w3.org/2000/svg" width="100" height="20">
<linearGradient id="a" x2="0" y2="100%">
  <stop offset="0" stop-color="#bbb" stop-opacity=".1"/>
  <stop offset="2" stop-opacity=".1"/>
</linearGradient>

<rect rx="3" width="60" height="20" fill="#555"/> <!-- Comment -->
<rect rx="3" x="60" width="40" height="20" fill="#4c1"/>

<path fill="#4c1" d="M58 0h4v20h-4z"/>

<rect rx="3" width="100" height="20" fill="url(#a)"/>
<g fill="#fff" text-anchor="middle" font-family="DejaVu Sans,Verdana,Geneva,sans-serif" font-size="11">
    <text x="30" y="14">{}</text>
    <text x="80" y="14">{}</text>
</g>
</svg>'''


parser = argparse.ArgumentParser(description='Use a summary.json file to generate a test coverage badge')
parser.add_argument('-i', "--input_summary_file", type=argparse.FileType('r'), required = True,
                    help='Summary file to use to generate the pass percentage badge')
parser.add_argument('-o', "--output_badge_file", type=argparse.FileType('w'), required = True,
                    help='Name of the badge to output')
parser.add_argument('-s', "--badge_string", type=str, required = True,
                    help='Name of the badge to output')



try:
   results = parser.parse_args()
except Exception as e:
   print(f"Error parsing arguments: {str(e)}")
   exit(1)

try:
   summary_info = json.loads(results.input_summary_file.read())
except Exception as e:
   print(f"Error loading {results.input_summary_file.name} JSON file: {str(e)}")
   sys.exit(1)

if 'summary' not in summary_info:
   print("Missing 'summary' key in {results.input_summary_file.name}")
   sys.exit(1)
elif 'PASS_RATE' not in summary_info['summary'] or 'TESTS_PASSED' not in summary_info['summary']:
   print(f"Missing PASS_RATE in 'summary' section of {results.input_summary_file.name}")
   sys.exit(1)
pass_percent = 100 * summary_info['summary']['PASS_RATE']


try:
   results.output_badge_file.write(RAW_BADGE_SVG.format(results.badge_string, "{:2.1f}%".format(pass_percent)))
except Exception as e:
   print(f"Error generating badge: {str(e)}")
   sys.exit(1)


print(f"Badge {results.output_badge_file.name} successfully generated!")
sys.exit(0)




import string
import uuid
import requests

from pydantic import BaseModel, validator, ValidationError
from datetime import datetime

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.story_tags import StoryTags

class Story(BaseModel, SecurityContentObject):
    # story spec
    name: str
    id: str
    version: int
    date: str
    author: str
    description: str
    narrative: str
    references: list
    tags: StoryTags

    # enrichments
    detection_names: list = None
    investigation_names: list = None
    baseline_names: list = None
    author_company: str = None
    author_name: str = None
    detections: list = None
    investigations: list = None
    
    
    @validator('name')
    def name_invalid_chars(cls, v):
        invalidChars = set(string.punctuation.replace("-", ""))
        if any(char in invalidChars for char in v):
            raise ValueError('invalid chars used in name: ' + v)
        return v

    @validator('id')
    def id_check(cls, v, values):
        try:
            uuid.UUID(str(v))
        except:
            raise ValueError('uuid is not valid: ' + values["name"])
        return v

    @validator('date')
    def date_valid(cls, v, values):
        try:
            datetime.strptime(v, "%Y-%m-%d")
        except:
            raise ValueError('date is not in format YYYY-MM-DD: ' + values["name"])
        return v

    @validator('description', 'narrative')
    def encode_error(cls, v, values, field):
        try:
            v.encode('ascii')
        except UnicodeEncodeError:
            raise ValueError('encoding error in ' + field.name + ': ' + values["name"])
        return v



import sys

from pydantic import ValidationError

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.input.yml_reader import YmlReader
from contentctl.objects.enums import SecurityContentType
from contentctl.objects.deployment import Deployment
from contentctl.objects.macro import Macro
from contentctl.objects.lookup import Lookup
from contentctl.objects.playbook import Playbook
from contentctl.objects.unit_test import UnitTest


class BasicBuilder():
    security_content_obj : SecurityContentObject


    def setObject(self, path: str, type: SecurityContentType) -> None:
        yml_dict = YmlReader.load_file(path)
        if type == SecurityContentType.deployments:
            if "alert_action" in yml_dict:
                alert_action_dict = yml_dict["alert_action"]
                for key in alert_action_dict.keys():
                    yml_dict[key] = yml_dict["alert_action"][key]
            try:
                self.security_content_obj = Deployment.parse_obj(yml_dict)
            except ValidationError as e:
                print('Validation Error for file ' + path)
                print(e)
                sys.exit(1)
        elif type == SecurityContentType.macros:
            try:
                self.security_content_obj = Macro.parse_obj(yml_dict)
            except ValidationError as e:
                print('Validation Error for file ' + path)
                print(e)
                sys.exit(1)
        elif type == SecurityContentType.lookups:
            try:
                self.security_content_obj = Lookup.parse_obj(yml_dict)
            except ValidationError as e:
                print('Validation Error for file ' + path)
                print(e)
                sys.exit(1)
        elif type == SecurityContentType.unit_tests:
            try:
                self.security_content_obj = UnitTest.parse_obj(yml_dict)
            except ValidationError as e:
                print('Validation Error for file ' + path)
                print(e)
                sys.exit(1)
    
    def reset(self) -> None:
        self.security_content_obj = None

    def getObject(self) -> SecurityContentObject:
        return self.security_content_obj


# Needed for a staticmethod to be able to return an instance of the class it belongs to
from __future__ import annotations


import validators
import pathlib
import yaml
import os
from pydantic import BaseModel, validator, root_validator, Extra, Field
from dataclasses import dataclass
from typing import Union
import docker
import docker.errors


from contentctl.objects.enums import (
    PostTestBehavior,
    DetectionTestingMode,
    DetectionTestingTargetInfrastructure,
)

from contentctl.objects.app import App
from contentctl.helper.utils import Utils


ALWAYS_PULL_REPO = False
PREVIOUSLY_ALLOCATED_PORTS: set[int] = set()

LOCAL_APP_DIR = pathlib.Path("apps")
CONTAINER_APP_DIR = pathlib.Path("/tmp/apps")


def getTestConfigFromYMLFile(path: pathlib.Path):
    try:
        with open(path, "r") as config_handle:
            cfg = yaml.safe_load(config_handle)
        return TestConfig.parse_obj(cfg)

    except Exception as e:
        print(f"Error loading test configuration file '{path}': {str(e)}")


class TestConfig(BaseModel, extra=Extra.forbid, validate_assignment=True):
    repo_path: str = Field(default=".", title="Path to the root of your app")
    repo_url: Union[str, None] = Field(
        default=None,
        title="HTTP(s) path to the repo for repo_path.  If this field is blank, it will be inferred from the repo",
    )
    mock: bool = Field(
        default=False,
        title="Whether or not to actually run the test, or just generate the app and test configuration files",
    )
    # main_branch: Union[str,None] = Field(default=None, title="Main branch of the repo, if applicable.")
    # test_branch: Union[str,None] = Field(default=None, title="Branch of the repo to be tested, if applicable.")
    # commit_hash: Union[str,None] = Field(default=None, title="Commit hash of the repo state to be tested, if applicable")
    target_infrastructure: DetectionTestingTargetInfrastructure = Field(
        default=DetectionTestingTargetInfrastructure.container,
        title=f"Control where testing should be launched.  Choose one of {DetectionTestingTargetInfrastructure._member_names_}",
    )
    full_image_path: str = Field(
        default="registry.hub.docker.com/splunk/splunk:latest",
        title="Full path to the container image to be used",
    )
    container_name: str = Field(
        default="splunk_contentctl_%d",
        title="Template to be used for naming the Splunk Test Containers which will be created",
    )
    post_test_behavior: PostTestBehavior = Field(
        default=PostTestBehavior.pause_on_failure,
        title=f"What to do after a test has completed.  Choose one of {PostTestBehavior._member_names_}",
    )
    mode: DetectionTestingMode = Field(
        default=DetectionTestingMode.all,
        title=f"Control which detections should be tested.  Choose one of {DetectionTestingMode._member_names_}",
    )
    detections_list: Union[list[str], None] = Field(
        default=None, title="List of paths to detections which should be tested"
    )
    num_containers: int = Field(
        default=1, title="Number of testing containers to start in parallel."
    )
    # pr_number: Union[int,None] = Field(default=None, title="The number of the PR to test")
    splunk_app_username: Union[str, None] = Field(
        default="admin", title="The name of the user for testing"
    )
    splunk_app_password: Union[str, None] = Field(
        default="password", title="Password for logging into Splunk Server"
    )
    splunkbase_username: Union[str, None] = Field(
        default=None,
        title="The username for logging into Splunkbase in case apps must be downloaded",
    )
    splunkbase_password: Union[str, None] = Field(
        default=None,
        title="The password for logging into Splunkbase in case apps must be downloaded",
    )
    apps: list[App] = Field(
        default=App.get_default_apps(),
        title="A list of all the apps to be installed on each container",
    )
    test_instance_address: str = Field(
        default="127.0.0.1",
        title="Domain name of IP address of Splunk server to be used for testing. Do NOT use a protocol, like http(s):// or 'localhost'",
    )

    hec_port: int = Field(default=8088, title="HTTP Event Collector Port")
    web_ui_port: int = Field(default=8000, title="Web UI Port")
    api_port: int = Field(default=8089, title="REST API Port")

    @validator("hec_port", "web_ui_port", "api_port", each_item=True)
    def validate_ports_range(cls, v):
        if v < 2:
            raise (
                ValueError(
                    f"Error, invalid Port number. Port must be between 2-65535: {v}"
                )
            )
        elif v > 65535:
            raise (
                ValueError(
                    f"Error, invalid Port number. Port must be between 2-65535: {v}"
                )
            )
        return v

    @validator("hec_port", "web_ui_port", "api_port", each_item=False)
    def validate_ports_overlap(cls, v):
        global PREVIOUSLY_ALLOCATED_PORTS
        if type(v) is not list:
            # Otherwise this throws error when we update a single field
            return v
        if len(set(v)) != len(v):
            raise (ValueError(f"Duplicate ports detected: [{v}]"))

        if PREVIOUSLY_ALLOCATED_PORTS.isdisjoint(v):
            PREVIOUSLY_ALLOCATED_PORTS = PREVIOUSLY_ALLOCATED_PORTS.union()
        else:
            raise (
                ValueError(
                    f"Duplicate ports detected: [{PREVIOUSLY_ALLOCATED_PORTS.intersection(v)}]"
                )
            )

        return v

    # Ensure that at least 1 of test_branch, commit_hash, and/or pr_number were passed.
    # Otherwise, what are we testing??
    # @root_validator(pre=False)
    # def ensure_there_is_something_to_test(cls, values):
    #     if 'test_branch' not in values and 'commit_hash' not in values and'pr_number' not in values:
    #         if 'mode' in values and values['mode'] == DetectionTestingMode.changes:
    #             raise(ValueError(f"Under mode [{DetectionTestingMode.changes}], 'test_branch', 'commit_hash', and/or 'pr_number' must be defined so that we know what to test."))

    #     return values

    # @validator('repo_path', always=True)
    # def validate_repo_path(cls,v):

    #     try:
    #         path = pathlib.Path(v)
    #     except Exception as e:
    #         raise(ValueError(f"Error, the provided path is is not a valid path: '{v}'"))

    #     try:
    #         r = git.Repo(path)
    #     except Exception as e:
    #         raise(ValueError(f"Error, the provided path is not a valid git repo: '{path}'"))

    #     try:

    #         if ALWAYS_PULL_REPO:
    #             r.remotes.origin.pull()
    #     except Exception as e:
    #         raise ValueError(f"Error pulling git repository {v}: {str(e)}")

    #     return v

    # @validator('repo_url', always=True)
    # def validate_repo_url(cls, v, values):
    #     Utils.check_required_fields('repo_url', values, ['repo_path'])

    #     #First try to get the value from the repo
    #     try:
    #         remote_url_from_repo = git.Repo(values['repo_path']).remotes.origin.url
    #     except Exception as e:
    #         raise(ValueError(f"Error reading remote_url from the repo located at {values['repo_path']}"))

    #     if v is not None and remote_url_from_repo != v:
    #         raise(ValueError(f"The url of the remote repo supplied in the config file {v} does not "\
    #                           f"match the value read from the repository at {values['repo_path']}, {remote_url_from_repo}"))

    #     if v is None:
    #         v = remote_url_from_repo

    #     #Ensure that the url is the proper format
    #     try:
    #         if bool(validators.url(v)) == False:
    #             raise(Exception)
    #     except:
    #         raise(ValueError(f"Error validating the repo_url. The url is not valid: {v}"))

    #     return v

    # @validator('main_branch', always=True)
    # def valid_main_branch(cls, v, values):
    #     Utils.check_required_fields('main_branch', values, ['repo_path', 'repo_url'])

    #     if v is None:
    #         print(f"main_branch is not supplied.  Inferring from '{values['repo_path']}'...",end='')

    #         main_branch = Utils.get_default_branch_name(values['repo_path'], values['repo_url'])
    #         print(f"main_branch name '{main_branch}' inferred'")
    #         #continue with the validation
    #         v = main_branch

    #     try:
    #         Utils.validate_git_branch_name(values['repo_path'],values['repo_url'], v)
    #     except Exception as e:
    #         raise ValueError(f"Error validating main_branch: {str(e)}")
    #     return v

    # @validator('test_branch', always=True)
    # def validate_test_branch(cls, v, values):
    #     Utils.check_required_fields('test_branch', values, ['repo_path', 'repo_url', 'main_branch'])
    #     if v is None:
    #         print(f"No test_branch provided, so we will default to using the main_branch '{values['main_branch']}'")
    #         return values['main_branch']
    #     try:
    #         Utils.validate_git_branch_name(values['repo_path'],values['repo_url'], v)
    #     except Exception as e:
    #         raise ValueError(f"Error validating test_branch: {str(e)}")
    #     return v

    # @validator('commit_hash', always=True)
    # def validate_commit_hash(cls, v, values):
    #     Utils.check_required_fields('commit_hash', values, ['repo_path', 'repo_url', 'test_branch'])

    #     try:
    #         #We can a hash with this function too
    #         Utils.validate_git_hash(values['repo_path'],values['repo_url'], v, values['test_branch'])
    #     except Exception as e:
    #         raise ValueError(f"Error validating commit_hash '{v}': {str(e)}")
    #     return v

    @validator("full_image_path", always=True)
    def validate_full_image_path(cls, v, values):
        if (
            values.get("target_infrastructure", None)
            == DetectionTestingTargetInfrastructure.server.value
        ):
            print(
                f"No need to validate target image path {v}, testing target is preconfigured server"
            )
            return v
        # This behavior may change if we start supporting local/offline containers and
        # the logic to build them
        if ":" not in v:
            raise (
                ValueError(
                    f"Error, the image_name {v} does not include a tag.  A tagged container MUST be included to ensure consistency when testing"
                )
            )

        # Check to make sure we have the latest version of the image
        # We have this as a wrapped, nested try/except because if we
        # encounter some error in trying to get the latest version, but
        # we do have some version already, we will allow the test to continue.
        # For example, this may occur if an image has been previously downloaded,
        # but the server no longer has internet connectivity and can't get the
        # image again. in this case, don't fail - continue with the test
        try:
            try:
                # connectivity to docker server is validated previously
                client = docker.from_env()
                print(
                    f"Getting the latest version of the container image: {v}...",
                    end="",
                    flush=True,
                )
                client.images.pull(v)
                print("done")
            except docker.errors.APIError as e:
                print("error")
                if e.is_client_error():
                    if "invalid reference format" in str(e.explanation):
                        simple_explanation = f"The format of the docker image reference is incorrect. Please use a valid image reference"
                    else:
                        simple_explanation = (
                            f"The most likely cause of this error is that the image/tag "
                            "does not exist or it is stored in a private repository and you are not logged in."
                        )

                elif e.is_server_error():
                    simple_explanation = (
                        f"The mostly likely cause is that the server cannot be reached. "
                        "Please ensure that the server hosting your docker image is available "
                        "and you have internet access, if required."
                    )

                else:
                    simple_explanation = f"Unable to pull image {v} for UNKNOWN reason. Please consult the detailed error below."

                verbose_explanation = e.explanation

                raise (
                    ValueError(
                        f"Error Pulling Docker Image '{v}'\n  - EXPLANATION: {simple_explanation} (full error text: '{verbose_explanation}'"
                    )
                )
            except Exception as e:
                print("error")
                raise (ValueError(f"Uknown error pulling Docker Image '{v}': {str(e)}"))

        except Exception as e:
            # There was some exception that prevented us from getting the latest version
            # of the image. However, if we already have it, use the current version and
            # down fully raise the exception - just use it
            client = docker.from_env()
            try:
                client.api.inspect_image(v)
                print(e)
                print(
                    f"We will default to using the version of the image {v} which has "
                    "already been downloaded to this machine. Please note that it may be out of date."
                )

            except Exception as e2:
                raise (
                    ValueError(
                        f"{str(e)}Image is not previously cached, so we could not use an old version."
                    )
                )

        return v

    # presumably the post test behavior is validated by the enum?
    # presumably the mode is validated by the enum?

    @validator("detections_list", always=True)
    def validate_detections_list(cls, v, values):

        Utils.check_required_fields("detections_list", values, ["mode", "repo_path"])
        # A detections list can only be provided if the mode is selected
        # otherwise, we must throw an error

        # First check the mode
        if values["mode"] != DetectionTestingMode.selected:
            if v is not None:
                # We intentionally raise an error even if the list is an empty list
                raise (
                    ValueError(
                        f"For Detection Testing Mode '{values['mode']}', "
                        f"'detections_list' MUST be none.  Instead, it was a list containing {len(v)} detections."
                    )
                )
            return v

        # Mode is DetectionTestingMode.selected - verify the paths of all the detections
        all_errors = []
        if v == None:
            raise (
                ValueError(
                    f"mode is '{DetectionTestingMode.selected}', but detections_list was not provided."
                )
            )
        for detection in v:
            try:
                full_path = os.path.join(values["repo_path"], detection)
                if not pathlib.Path(full_path).exists():
                    all_errors.append(full_path)
            except Exception as e:
                all_errors.append(
                    f"Unexpected error validating path '{detection}': {str(e)}"
                )
        if len(all_errors):
            joined_errors = "\n\t".join(all_errors)
            raise (
                ValueError(
                    f"Paths to the following detections in 'detections_list' "
                    f"were invalid: \n\t{joined_errors}"
                )
            )

        return v

    @validator("num_containers", always=True)
    def validate_num_containers(cls, v):
        MAX_RECOMMENDED_CONTAINERS_BEFORE_WARNING = 2
        if v < 1:
            raise (
                ValueError(
                    f"Error validating num_containers. Test must be run with at least 1 container, not {v}"
                )
            )
        if v > MAX_RECOMMENDED_CONTAINERS_BEFORE_WARNING:
            print(
                f"You requested to run with [{v}] containers which may use a very large amount of resources "
                "as they all run in parallel.  The maximum suggested number of parallel containers is "
                f"[{MAX_RECOMMENDED_CONTAINERS_BEFORE_WARNING}].  We will do what you asked, but be warned!"
            )
        return v

    # @validator('pr_number', always=True)
    # def validate_pr_number(cls, v, values):
    #     Utils.check_required_fields('pr_number', values, ['repo_path', 'commit_hash'])

    #     if v == None:
    #         return v

    #     hash = Utils.validate_git_pull_request(values['repo_path'], v)

    #     #Ensure that the hash is equal to the one in the config file, if it exists.
    #     if values['commit_hash'] is None:
    #         values['commit_hash'] = hash
    #     else:
    #         if values['commit_hash'] != hash:
    #             raise(ValueError(f"commit_hash specified in configuration was {values['commit_hash']}, but commit_hash"\
    #                              f" from pr_number {v} was {hash}. These must match.  If you're testing"\
    #                              " a PR, you probably do NOT want to provide the commit_hash in the configuration file "\
    #                              "and always want to test the head of the PR. This will be done automatically if you do "\
    #                              "not provide the commit_hash."))

    #     return v

    @validator("splunk_app_password", always=True)
    def validate_splunk_app_password(cls, v):
        if v == None:
            # No app password was provided, so generate one
            v = Utils.get_random_password()
        else:
            MIN_PASSWORD_LENGTH = 6
            if len(v) < MIN_PASSWORD_LENGTH:
                raise (
                    ValueError(
                        f"Password is less than {MIN_PASSWORD_LENGTH} characters long. This password is extremely weak, please change it."
                    )
                )
        return v

    @validator("splunkbase_username", always=True)
    def validate_splunkbase_username(cls, v):
        return v

    @validator("splunkbase_password", always=True)
    def validate_splunkbase_password(cls, v, values):
        Utils.check_required_fields("repo_url", values, ["splunkbase_username"])
        if values["splunkbase_username"] == None:
            return v
        elif (v == None and values["splunkbase_username"] != None) or (
            v != None and values["splunkbase_username"] == None
        ):
            raise (
                ValueError(
                    "splunkbase_username OR splunkbase_password "
                    "was provided, but not both.  You must provide"
                    " neither of these value or both, but not just "
                    "1 of them"
                )
            )

        else:
            return v

    @validator("apps")
    def validate_apps(cls, v, values):
        Utils.check_required_fields(
            "repo_url", values, ["splunkbase_username", "splunkbase_password"]
        )

        app_errors = []

        # ensure that the splunkbase username and password are provided
        username = values["splunkbase_username"]
        password = values["splunkbase_password"]
        app_directory = pathlib.Path(values["repo_path"]) / LOCAL_APP_DIR
        try:
            os.makedirs(LOCAL_APP_DIR, exist_ok=True)
        except Exception as e:
            raise (
                Exception(f"Error: When trying to create {CONTAINER_APP_DIR}: {str(e)}")
            )

        for app in v:
            try:
                app.configure_app_source_for_container(
                    username, password, app_directory, CONTAINER_APP_DIR
                )
            except Exception as e:
                error_string = f"Unable to prepare app '{app.title}': {str(e)}"
                app_errors.append(error_string)

        if len(app_errors) != 0:
            error_string = "\n\t".join(app_errors)
            raise (ValueError(f"Error preparing apps to install:\n\t{error_string}"))

        return v

    @validator("target_infrastructure", always=True)
    def validate_target_infrastructure(cls, v, values):
        if v == DetectionTestingTargetInfrastructure.server.value:
            # No need to validate that the docker client is available
            return v
        elif v == DetectionTestingTargetInfrastructure.container.value:
            if values["mock"] is False:
                # we need to make sure we can actually get the docker client from the environment
                try:
                    docker.client.from_env()
                except Exception as e:
                    raise (
                        Exception(
                            f"Error, failed to get docker client.  Is Docker Installed and running "
                            f"and are docker environment variables set properly? Error:\n\t{str(e)}"
                        )
                    )
        return v

    @validator("test_instance_address", always=True)
    def validate_test_instance_address(cls, v, values):
        try:
            if v.startswith("http"):
                raise (Exception("should not begin with http"))
            is_ipv4 = validators.ipv4(v)
            if bool(is_ipv4):
                return v
            is_domain_name = validators.domain(v)
            if bool(is_domain_name):
                import socket

                try:
                    socket.gethostbyname(v)
                    return v
                except Exception as e:
                    pass
                raise (Exception("DNS Lookup failed"))
            raise (Exception(f"not an IPV4 address or a domain name"))
        except Exception as e:
            raise (
                Exception(
                    f"Error, failed to validate test_instance_address '{v}': {str(e)}"
                )
            )



import re
import sys

from pydantic import ValidationError

from contentctl.objects.investigation import Investigation
from contentctl.input.yml_reader import YmlReader
from contentctl.objects.enums import SecurityContentType


class InvestigationBuilder():
    investigation: Investigation

    def setObject(self, path: str) -> None:
        yml_dict = YmlReader.load_file(path)
        try:
            self.investigation = Investigation.parse_obj(yml_dict)
        except ValidationError as e:
            print('Validation Error for file ' + path)
            print(e)
            sys.exit(1)

    def reset(self) -> None:
        self.investigation = None


    def getObject(self) -> Investigation:
        return self.investigation


    def addInputs(self) -> None:
        pattern = r"\$([^\s.]*)\$"
        inputs = []

        for input in re.findall(pattern, self.investigation.search):
            inputs.append(input)

        self.investigation.inputs = inputs

    
    def addLowercaseName(self) -> None:
        self.investigation.lowercase_name = self.investigation.name.replace(' ', '_').replace('-','_').replace('.','_').replace('/','_').lower().replace(' ', '_').replace('-','_').replace('.','_').replace('/','_').lower()


from pydantic import BaseModel, root_validator, validator


from typing import Union
from datetime import timedelta
from splunklib.data import Record
from contentctl.objects.test_config import TestConfig
from contentctl.helper.utils import Utils

FORCE_TEST_FAILURE_FOR_MISSING_OBSERVABLE = False

SID_TEMPLATE = "{server}:{web_port}/en-US/app/search/search?sid={sid}"


class UnitTestResult(BaseModel):
    job_content: Union[Record, None] = None
    missing_observables: list[str] = []
    sid_link: Union[None, str] = None
    message: Union[None, str] = None
    exception: bool = False
    success: bool = False
    duration: float = 0

    class Config:
        validate_assignment = True

    def get_summary_dict(
        self,
        model_fields: list[str] = ["success", "exception", "message", "sid_link"],
        job_fields: list[str] = ["search", "resultCount", "runDuration"],
    ) -> dict:
        results_dict = {}
        for field in model_fields:
            value = getattr(self, field)
            results_dict[field] = getattr(self, field)

        for field in job_fields:
            if self.job_content is not None:
                value = self.job_content.get(field, None)
                if field == "runDuration":
                    try:
                        value = Utils.getFixedWidth(float(value), 3)
                    except Exception as e:
                        value = Utils.getFixedWidth(0, 3)
                results_dict[field] = value
            else:
                results_dict[field] = None

        return results_dict

    def set_job_content(
        self,
        content: Union[Record, None, Exception],
        config: TestConfig,
        success: bool = False,
        duration: float = 0,
    ):
        self.duration = round(duration, 2)
        if isinstance(content, Record):
            self.job_content = content
            self.success = success
            if success:
                self.message = "TEST PASSED"
            else:
                self.message = "TEST FAILED"
            self.exception = False

            if not config.test_instance_address.startswith("http://"):
                sid_template = f"http://{SID_TEMPLATE}"
            else:
                sid_template = SID_TEMPLATE
            self.sid_link = sid_template.format(
                server=config.test_instance_address,
                web_port=config.web_ui_port,
                sid=content.get("sid", None),
            )

        elif isinstance(content, Exception):
            self.job_content = None
            self.success = False
            self.exception = True
            self.message = f"Error during test: {str(content)}"

        elif content is None:
            self.job_content = None
            self.success = False
            self.exception = True
            self.message = f"Error during test: unable to run test"

        else:
            msg = f"Error: Unknown type for content in UnitTestResult: {type(content)}"
            print(msg)
            self.job_content = None
            self.success = False
            self.exception = True
            self.message = f"Error during test - unable to run test {msg}"
        return self.success

    """
    def get_summary(self, test_name: str, verbose=False) -> str:
        lines: list[str] = []
        lines.append(f"SEARCH NAME        : '{test_name}'")
        if verbose or self.determine_success() == False:
            lines.append(f"SEARCH             : {self.get_search()}")
            lines.append(f"SUCCESS            : {self.determine_success()}")
            if self.exception is True:
                lines.append(f"EXCEPTION          : {self.exception}")
            if self.message is not None:
                lines.append(f"MESSAGE            : {self.message}")
        else:
            lines.append(f"SUCCESS            : {self.determine_success()}")
        if len(self.missing_observables) > 0:
            lines.append(f"MISSING OBSERVABLES: {self.missing_observables}")

        return "\n\t".join(lines)

    def get_search(self) -> str:
        if self.job_content is not None:
            return self.job_content.get(
                "search", "NO SEARCH FOUND - JOB MISSING SEARCH FIELD"
            )
        return "NO SEARCH FOUND - JOB IS EMPTY"

    def add_message(self, message: str):
        if self.message is None:
            self.message = message
        else:
            self.message += f"\n{message}"

    @root_validator(pre=False)
    def update_success(cls, values):
        if values["job_content"] is None:
            values["exception"] = True
            values["success"] = False
            if values["message"] is None:
                # If the message has not been overridden, then put in a default
                values["message"] = "Job Content was None - unknown failure reason"
            # Otherwise, a message has been passed so don't overwrite it
            return values

        if "messages" in values["job_content"]:
            fatal_or_error = False
            all_messages = values["job_content"]["messages"]
            unique_messages = set()
            for level, level_messages in all_messages.items():
                if level in ["info"]:
                    # we will skip any info messages
                    continue
                elif level in ["fatal", "error"]:
                    for msg in level_messages:
                        # These error indicate a failure - the search was
                        # not successful. They are important for debugging,
                        # so we will pass them to the user.
                        # They also represent a an error during the test
                        values["logic"] = False
                        values["success"] = False
                        values["exception"] = True
                        unique_messages.add(msg)
                        fatal_or_error = True
                else:
                    unknown_messages_as_single_string = "\n".join(level_messages)
                    unique_messages.add(unknown_messages_as_single_string)

            if len(unique_messages) == 0:
                values["message"] = None  # No messages

            else:
                # Merge all those messages together
                values["message"] = "\n".join(unique_messages)

            if fatal_or_error:
                return values

        # Can there still be a success even if there was an error/fatal message above? Probably not?
        if (
            "resultCount" in values["job_content"]
            and int(values["job_content"]["resultCount"]) == 1
        ):
            # in the future we probably want other metrics, about noise or others, here
            values["logic"] = True
            values["success"] = True

        elif (
            "resultCount" in values["job_content"]
            and int(values["job_content"]["resultCount"]) != 1
        ):
            values["logic"] = False
            values["success"] = False

        else:
            raise (Exception("Result created with indeterminate success."))

        return values

    def update_missing_observables(self, missing_observables: set[str]):
        self.missing_observables = list(missing_observables)
        self.success = self.determine_success()

    def determine_success(self) -> bool:
        # values_dict = self.update_success(self.__dict__)
        # self.exception = values_dict['exception']
        # self.success = values_dict['success']
        return self.success

    def get_job_field(self, fieldName: str):
        if self.job_content is None:
            # return f"FIELD NAME {fieldName} does not exist in Job Content because Job Content is NONE"
            return None
        return self.job_content.get(fieldName, None)

    def get_time(self) -> timedelta:
        if self.job_content is None:
            return timedelta(0)
        elif "runDuration" in self.job_content:
            duration = str(self.job_content["runDuration"])
            return timedelta(float(duration))
        else:
            raise (Exception("runDuration missing from job."))
    """



import datetime
import os
from xmlrpc.client import APPLICATION_ERROR
from jinja2 import Environment, FileSystemLoader, StrictUndefined

from contentctl.objects.security_content_object import SecurityContentObject

class ConfWriter():

    @staticmethod
    def writeConfFileHeader(output_path : str) -> None:
        utc_time = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
        j2_env = Environment(
            loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')), 
            trim_blocks=True)

        template = j2_env.get_template('header.j2')
        output = template.render(time=utc_time)
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'w') as f:
            output = output.encode('ascii', 'ignore').decode('ascii')
            f.write(output)


    @staticmethod
    def writeConfFileHeaderEmpty(output_path : str) -> None:
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'w') as f:
            f.write('')


    @staticmethod
    def writeConfFile(template_name : str, output_path : str, objects : list, app_name:str) -> None:
        def custom_jinja2_enrichment_filter(string, object):
            customized_string = string

            for key in dir(object):
                if type(key) is not str:
                    key = key.decode()
                if not key.startswith('__') and not key == "_abc_impl" and not callable(getattr(object, key)):
                    if hasattr(object, key):
                        customized_string = customized_string.replace("%" + key + "%", str(getattr(object, key)))

            for key in dir(object.tags):
                if type(key) is not str:
                    key = key.decode()
                if not key.startswith('__') and not key == "_abc_impl" and not callable(getattr(object.tags, key)):
                    if hasattr(object.tags, key):
                        customized_string = customized_string.replace("%" + key + "%", str(getattr(object.tags, key)))

            return customized_string

        j2_env = Environment(
            loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')), 
            trim_blocks=True,
            undefined=StrictUndefined)

        j2_env.filters['custom_jinja2_enrichment_filter'] = custom_jinja2_enrichment_filter
        template = j2_env.get_template(template_name)
        output = template.render(objects=objects, APP_NAME=app_name)
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'a') as f:
            output = output.encode('ascii', 'ignore').decode('ascii')
            f.write(output)




import os

from jinja2 import Environment, FileSystemLoader


class JinjaWriter:

    @staticmethod
    def writeObjectsList(template_name : str, output_path : str, objects : list) -> None:

        j2_env = Environment(
            loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')), 
            trim_blocks=False)

        template = j2_env.get_template(template_name)
        output = template.render(objects=objects)
        with open(output_path, 'w') as f:
            output = output.encode('ascii', 'ignore').decode('ascii')
            f.write(output)


    @staticmethod
    def writeObject(template_name : str, output_path : str, object : dict) -> None:

        j2_env = Environment(
            loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')), 
            trim_blocks=False)

        template = j2_env.get_template(template_name)
        output = template.render(object=object)
        with open(output_path, 'w') as f:
            output = output.encode('ascii', 'ignore').decode('ascii')
            f.write(output)




SES_CONTEXT_MAPPING = {
    "Unknown": 0,
    "Source:Endpoint": 10,
    "Source:AD": 11,
    "Source:Firewall": 12,
    "Source:Application Log": 13,
    "Source:IPS": 14,
    "Source:Cloud Data": 15,
    "Source:Correlation": 16,
    "Source:Printer": 17,
    "Source:Badge": 18,
    "Scope:Internal": 20,
    "Scope:External": 21,
    "Scope:Inbound": 22,
    "Scope:Outbound": 23,
    "Scope:Local": 24,
    "Scope:Network": 25,
    "Outcome:Blocked": 30,
    "Outcome:Allowed": 31,
    "Stage:Recon": 40,
    "Stage:Initial Access": 41,
    "Stage:Execution": 42,
    "Stage:Persistence": 43,
    "Stage:Privilege Escalation": 44,
    "Stage:Defense Evasion": 45,
    "Stage:Credential Access": 46,
    "Stage:Discovery": 47,
    "Stage:Lateral Movement": 48,
    "Stage:Collection": 49,
    "Stage:Exfiltration": 50,
    "Stage:Command And Control": 51,
    "Consequence:Infection": 60,
    "Consequence:Reduced Visibility": 61,
    "Consequence:Data Destruction": 62,
    "Consequence:Denial Of Service": 63,
    "Consequence:Loss Of Control": 64,
    "Rares:Rare User": 70,
    "Rares:Rare Process": 71,
    "Rares:Rare Device": 72,
    "Rares:Rare Domain": 73,
    "Rares:Rare Network": 74,
    "Rares:Rare Location": 75,
    "Other:Peer Group": 80,
    "Other:Brute Force": 81,
    "Other:Policy Violation": 82,
    "Other:Threat Intelligence": 83,
    "Other:Flight Risk": 84,
    "Other:Removable Storage": 85
}

SES_KILL_CHAIN_MAPPINGS = {
    "Unknown": 0,
    "Reconnaissance": 1,
    "Weaponization": 2,
    "Delivery": 3,
    "Exploitation": 4,
    "Installation": 5,
    "Command & Control": 6,
    "Actions on Objectives": 7
}

SES_OBSERVABLE_ROLE_MAPPING = {
    "Other": -1,
    "Unknown": 0,
    "Actor": 1,
    "Target": 2,
    "Attacker": 3,
    "Victim": 4,
    "Parent Process": 5,
    "Child Process": 6,
    "Known Bad": 7,
    "Data Loss": 8,
    "Observer": 9
}

SES_OBSERVABLE_TYPE_MAPPING = {
    "Other": -1,
    "Unknown": 0,
    "Device": 1,
    "Container": 2,
    "Endpoint": 3,
    "Hostname": 4,
    "IP Address": 5,
    "User": 6,
    "Username": 7,
    "Email": 8,
    "Email Address": 9,
    "URL": 10,
    "URL Domain": 11,
    "File": 12,
    "File Name": 13,
    "File Hash": 14,
    "Process": 15,
    "Process Name": 16,
    "Location": 17
}


import os


from contentctl.output.json_writer import JsonWriter
from contentctl.objects.enums import SecurityContentType


class ApiJsonOutput():

    def writeObjects(self, objects: list, output_path: str, type: SecurityContentType = None) -> None:
        if type == SecurityContentType.detections:
            obj_array = []
            for detection in objects:
                obj_array.append(detection.dict(exclude_none=True, 
                    exclude =
                    {
                        "deprecated": True,
                        "experimental": True,
                        "annotations": True,
                        "risk": True,
                        "playbooks": True,
                        "baselines": True,
                        "mappings": True,
                        "test": True,
                        "deployment": True
                    }
                ))
            
            JsonWriter.writeJsonObject(os.path.join(output_path, 'detections.json'), {'detections': obj_array })
        
        elif type == SecurityContentType.stories:
            obj_array = []
            for story in objects:
                obj_array.append(story.dict(exclude_none=True,
                    exclude =
                    {
                        "investigations": True
                    }
                ))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'stories.json'), {'stories': obj_array })

        elif type == SecurityContentType.baselines:
            obj_array = []
            for baseline in objects:
                obj_array.append(baseline.dict(
                    exclude =
                    {
                        "deployment": True
                    }
                ))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'baselines.json'), {'baselines': obj_array })

        elif type == SecurityContentType.investigations:
            obj_array = []
            for investigation in objects:
                obj_array.append(investigation.dict(exclude_none=True))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'response_tasks.json'), {'response_tasks': obj_array })
        
        elif type == SecurityContentType.lookups:
            obj_array = []
            for lookup in objects:
                obj_array.append(lookup.dict(exclude_none=True))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'lookups.json'), {'lookups': obj_array })

        elif type == SecurityContentType.macros:      
            obj_array = []
            for macro in objects:
                obj_array.append(macro.dict(exclude_none=True))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'macros.json'), {'macros': obj_array })

        elif type == SecurityContentType.deployments:
            obj_array = []
            for deployment in objects:
                obj_array.append(deployment.dict(exclude_none=True))

            JsonWriter.writeJsonObject(os.path.join(output_path, 'deployments.json'), {'deployments': obj_array })




import os
import collections
import sys


from contentctl.input.yml_reader import YmlReader
from contentctl.objects.config import Config


class ConfigHandler:

    @classmethod
    def read_config(cls, config_path: str) -> Config:
        try:
            yml_dict = YmlReader.load_file(config_path)
        except:
            print("ERROR: no contentctl.yml found in given path")
            sys.exit(1)

        try: 
            config = Config.parse_obj(yml_dict)
        except Exception as e:
            print(e)
            sys.exit(1)

        return config

          
 
 


import uuid
import string
import requests
import time
import sys

from pydantic import BaseModel, validator, root_validator, Extra
from dataclasses import dataclass
from datetime import datetime, timedelta


from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.enums import AnalyticsType
from contentctl.objects.enums import DataModel
from contentctl.objects.detection_tags import DetectionTags
from contentctl.objects.config import ConfigDetectionConfiguration
from contentctl.objects.unit_test import UnitTest
from contentctl.objects.macro import Macro
from contentctl.objects.lookup import Lookup
from contentctl.objects.baseline import Baseline
from contentctl.objects.playbook import Playbook
from contentctl.helper.link_validator import LinkValidator


from typing import Union


class Detection(BaseModel, SecurityContentObject):
    # detection spec
    name: str
    id: str
    version: int
    date: str
    author: str
    type: str
    datamodel: list
    description: str
    search: str
    how_to_implement: str
    known_false_positives: str
    check_references: bool = (
        False  # Validation is done in order, this field must be defined first
    )
    references: list
    tags: DetectionTags

    # enrichments
    deprecated: bool = None
    experimental: bool = None
    deployment: ConfigDetectionConfiguration = None
    annotations: dict = None
    risk: list = None
    playbooks: list[Playbook] = None
    baselines: list[Baseline] = None
    mappings: dict = None
    test: UnitTest = None
    macros: list[Macro] = None
    lookups: list[Lookup] = None
    cve_enrichment: list = None
    splunk_app_enrichment: list = None
    file_path: str = None
    source: str = None
    nes_fields: str = None
    providing_technologies: list = None

    # @validator('name')
    # def name_max_length(cls, v, values):
    #     if len(v) > 67:
    #         raise ValueError('name is longer then 67 chars: ' + v)
    #     return v

    @validator("name")
    def name_invalid_chars(cls, v):
        invalidChars = set(string.punctuation.replace("-", ""))
        if any(char in invalidChars for char in v):
            raise ValueError("invalid chars used in name: " + v)
        return v

    @validator("id")
    def id_check(cls, v, values):
        try:
            uuid.UUID(str(v))
        except:
            raise ValueError("uuid is not valid: " + values["name"])
        return v

    @validator("date")
    def date_valid(cls, v, values):
        try:
            datetime.strptime(v, "%Y-%m-%d")
        except:
            raise ValueError("date is not in format YYYY-MM-DD: " + values["name"])
        return v

    @validator("type")
    def type_valid(cls, v, values):
        if v.lower() not in [el.name.lower() for el in AnalyticsType]:
            raise ValueError("not valid analytics type: " + values["name"])
        return v

    @validator("datamodel")
    def datamodel_valid(cls, v, values):
        for datamodel in v:
            if datamodel not in [el.name for el in DataModel]:
                raise ValueError("not valid data model: " + values["name"])
        return v

    @validator("description", "how_to_implement")
    def encode_error(cls, v, values, field):
        try:
            v.encode("ascii")
        except UnicodeEncodeError:
            raise ValueError("encoding error in " + field.name + ": " + values["name"])
        return v

    @root_validator
    def search_validation(cls, values):
        if "ssa_" not in values["file_path"]:
            if not "_filter" in values["search"]:
                raise ValueError("filter macro missing in: " + values["name"])
            if any(
                x in values["search"]
                for x in ["eventtype=", "sourcetype=", " source=", "index="]
            ):
                if not "index=_internal" in values["search"]:
                    raise ValueError(
                        "Use source macro instead of eventtype, sourcetype, source or index in detection: "
                        + values["name"]
                    )
        return values

    @root_validator
    def name_max_length(cls, values):
        # Check max length only for ESCU searches, SSA does not have that constraint
        if "ssa_" not in values["file_path"]:
            if len(values["name"]) > 67:
                raise ValueError("name is longer then 67 chars: " + values["name"])
        return values

    # disable it because of performance reasons
    # @validator('references')
    # def references_check(cls, v, values):
    #     LinkValidator.check_references(v, values["name"])
    #     return v

    @validator("search")
    def search_validate(cls, v, values):
        # write search validator
        return v

    def all_tests_successful(self) -> bool:
        if self.test is None or len(self.test.tests) == 0:
            return False
        for test in self.test.tests:
            if test.result is None or test.result.success == False:
                return False
        return True

    def get_summary(
        self,
        detection_fields: list[str] = ["name", "search"],
        test_model_fields: list[str] = ["success", "message"],
        test_job_fields: list[str] = ["resultCount", "runDuration"],
    ) -> dict:
        summary_dict = {}
        for field in detection_fields:
            summary_dict[field] = getattr(self, field)
        summary_dict["success"] = self.all_tests_successful()
        summary_dict["tests"] = []
        if self.test is not None:
            for test in self.test.tests:
                result: dict[str, Union[str, bool]] = {"name": test.name}
                if test.result is not None:
                    result.update(
                        test.result.get_summary_dict(
                            model_fields=test_model_fields,
                            job_fields=test_job_fields,
                        )
                    )
                else:
                    result["success"] = False
                    result["message"] = "RESULT WAS NONE"

                summary_dict["tests"].append(result)

        return summary_dict





from dataclasses import dataclass

from contentctl.input.new_content_generator import NewContentGenerator, NewContentGeneratorInputDto, NewContentGeneratorOutputDto
from contentctl.output.new_content_yml_output import NewContentYmlOutput


@dataclass(frozen=True)
class NewContentInputDto:
    new_content_generator_input_dto: NewContentGeneratorInputDto
    output_path: str


class NewContent:

    def execute(self, input_dto: NewContentInputDto) -> None:
        new_content_generator_output_dto = NewContentGeneratorOutputDto({})
        new_content_generator = NewContentGenerator(new_content_generator_output_dto)
        new_content_generator.execute(input_dto.new_content_generator_input_dto)

        new_content_yml_output = NewContentYmlOutput(input_dto.output_path)
        new_content_yml_output.writeObjectNewContent(new_content_generator_output_dto.obj, input_dto.new_content_generator_input_dto.type)





from pydantic import BaseModel, validator, ValidationError
from typing import Union

class UnitTestBaseline(BaseModel):
    name: str
    file: str
    pass_condition: str
    earliest_time: Union[str,None] = None
    latest_time: Union[str,None] = None


from pydantic import BaseModel, validator, ValidationError


class MitreAttackEnrichment(BaseModel):
    mitre_attack_id: str
    mitre_attack_technique: str
    mitre_attack_tactics: list
    mitre_attack_groups: list




import uuid
import string

from pydantic import BaseModel, validator, ValidationError
from datetime import datetime

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.deployment_scheduling import DeploymentScheduling
from contentctl.objects.deployment_email import DeploymentEmail
from contentctl.objects.deployment_notable import DeploymentNotable
from contentctl.objects.deployment_rba import DeploymentRBA
from contentctl.objects.deployment_slack import DeploymentSlack
from contentctl.objects.deployment_phantom import DeploymentPhantom

class Deployment(BaseModel, SecurityContentObject):
    name: str
    id: str
    date: str
    author: str
    description: str
    scheduling: DeploymentScheduling = None
    email: DeploymentEmail = None
    notable: DeploymentNotable = None
    rba: DeploymentRBA = None
    slack: DeploymentSlack = None
    phantom: DeploymentPhantom = None
    tags: dict
    

    @validator('name')
    def name_invalid_chars(cls, v):
        invalidChars = set(string.punctuation.replace("-", ""))
        if any(char in invalidChars for char in v):
            raise ValueError('invalid chars used in name: ' + v)
        return v

    @validator('id')
    def id_check(cls, v, values):
        try:
            uuid.UUID(str(v))
        except:
            raise ValueError('uuid is not valid: ' + values["name"])
        return v

    @validator('date')
    def date_valid(cls, v, values):
        try:
            datetime.strptime(v, "%Y-%m-%d")
        except:
            raise ValueError('date is not in format YYYY-MM-DD: ' + values["name"])
        return v

    @validator('description')
    def encode_error(cls, v, values, field):
        try:
            v.encode('ascii')
        except UnicodeEncodeError:
            raise ValueError('encoding error in ' + field.name + ': ' + values["name"])
        return v



from pydantic import BaseModel, validator, ValidationError


class DeploymentPhantom(BaseModel):
    cam_workers : str
    label : str
    phantom_server : str
    sensitivity : str
    severity : str


import os


from contentctl.objects.enums import SecurityContentType
from contentctl.output.jinja_writer import JinjaWriter


class SvgOutput():

    def writeObjects(self, objects: list, output_path: str, type: SecurityContentType = None) -> None:
        
        detections_tmp = objects
        detection_without_test = 0    
      
        detections = []
        obj = dict()

        for detection in detections_tmp:
            if not detection.deprecated:
                detections.append(detection)

                if not detection.test and not detection.experimental:
                    detection_without_test = detection_without_test + 1


        obj['count'] = len(detections) 
        obj['coverage'] = (obj['count'] - detection_without_test)/obj['count']
        obj['coverage'] = "{:.0%}".format(obj['coverage'])

        JinjaWriter.writeObject('detection_count.j2', os.path.join(output_path, 'detection_count.svg'), obj)
        JinjaWriter.writeObject('detection_coverage.j2', os.path.join(output_path, 'detection_coverage.svg'), obj)




from pydantic import BaseModel
import abc
from typing import Callable
from contentctl.objects.test_config import TestConfig
from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingManagerOutputDto,
)

from contentctl.actions.detection_testing.views.DetectionTestingView import (
    DetectionTestingView,
)
import pathlib
import yaml

OUTPUT_FOLDER = "test_results"
OUTPUT_FILENAME = "summary.yml"


class DetectionTestingViewFile(DetectionTestingView):
    output_folder: str = OUTPUT_FOLDER
    output_filename: str = OUTPUT_FILENAME

    def getOutputFilePath(self) -> pathlib.Path:

        folder_path = pathlib.Path(self.config.repo_path) / self.output_folder
        output_file = folder_path / self.output_filename

        return output_file

    def setup(self):
        pass

    def stop(self):
        folder_path = pathlib.Path(self.config.repo_path) / OUTPUT_FOLDER
        output_file = self.getOutputFilePath()

        folder_path.mkdir(parents=True, exist_ok=True)

        result_dict = self.getSummaryObject()

        # use the yaml writer class
        with open(output_file, "w") as res:
            res.write(yaml.dump(result_dict))

    def showStatus(self, interval: int = 60):
        pass

    def showResults(self):
        pass

    def createReport(self):
        pass




from pydantic import BaseModel, validator, ValidationError



class BaselineTags(BaseModel):
    analytic_story: list
    deployments: list = None
    detections: list
    product: list
    required_fields: list
    security_domain: str


    @validator('product')
    def tags_product(cls, v, values):
        valid_products = [
            "Splunk Enterprise", "Splunk Enterprise Security", "Splunk Cloud",
            "Splunk Security Analytics for AWS", "Splunk Behavioral Analytics"
        ]

        for value in v:
            if value not in valid_products:
                raise ValueError('product is not valid for ' + values['name'] + '. valid products are ' + str(valid_products))
        return v


import requests
import xmltodict
import json

SPLUNKBASE_API_URL = "https://apps.splunk.com/api/apps/entriesbyid/"


class SplunkAppEnrichment():

    @classmethod
    def enrich_splunk_app(self, splunk_ta: str) -> dict:
        appurl = SPLUNKBASE_API_URL + splunk_ta
        splunk_app_enriched = dict()
        try:
            response = requests.get(appurl)
            response_dict = xmltodict.parse(response.content)
            # check if list since data changes depending on answer
            url, results = self._parse_splunkbase_response(response_dict)
            # grab the app name
            for i in results:
                if i['@name'] == 'appName':
                    splunk_app_enriched['name'] = i['#text']
            # grab out the splunkbase url  
            if 'entriesbyid' in url:
                response = requests.get(url)
                response_dict = xmltodict.parse(response.content)
                #print(json.dumps(response_dict, indent=2))
                url, results = self._parse_splunkbase_response(response_dict)
                # chop the url so we grab the splunkbase portion but not direct download
                splunk_app_enriched['url'] = url.rsplit('/', 4)[0]
        except requests.exceptions.ConnectionError as connErr:
            # there was a connection error lets just capture the name
            splunk_app_enriched['name'] = splunk_ta
            splunk_app_enriched['url'] = ''

        return splunk_app_enriched

    def _parse_splunkbase_response(response_dict):
        if isinstance(response_dict['feed']['entry'], list):
            url = response_dict['feed']['entry'][0]['link']['@href']
            results = response_dict['feed']['entry'][0]['content']['s:dict']['s:key']
        else:
            url = response_dict['feed']['entry']['link']['@href']
            results = response_dict['feed']['entry']['content']['s:dict']['s:key']
        return url, results




import requests
import urllib3, urllib3.exceptions


DEFAULT_USER_AGENT_STRING = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.41 Safari/537.36"
ALLOWED_HTTP_CODES = [200]
TIMEOUT_SEC = 15
ALLOW_REDIRECT = True
VERIFY_SSL = False


class LinkValidator:

    @staticmethod
    def check_references(references: list, name: str):
        for reference in references:
            try:
                get = requests.get(
                    reference, 
                    timeout=TIMEOUT_SEC, 
                    headers = {"User-Agent": DEFAULT_USER_AGENT_STRING}, 
                    allow_redirects=ALLOW_REDIRECT, 
                    verify=VERIFY_SSL
                )

                if get.status_code not in ALLOWED_HTTP_CODES:
                    print(get.status_code)
                    raise Exception(f"Reference Link Failed: {reference} for object {name}")
            

            except Exception as e:
                raise Exception(f"Reference Link Failed: {reference} for object {name}")


import sys
import shutil
import os

from dataclasses import dataclass

from contentctl.objects.enums import SecurityContentProduct, SecurityContentType
from contentctl.input.director import Director, DirectorInputDto, DirectorOutputDto
from contentctl.output.conf_output import ConfOutput
from contentctl.output.ba_yml_output import BAYmlOutput
from contentctl.output.api_json_output import ApiJsonOutput


@dataclass(frozen=True)
class GenerateInputDto:
    director_input_dto: DirectorInputDto


class Generate:

    def execute(self, input_dto: GenerateInputDto) -> DirectorOutputDto:
        director_output_dto = DirectorOutputDto([],[],[],[],[],[],[],[])
        director = Director(director_output_dto)
        director.execute(input_dto.director_input_dto)

        if input_dto.director_input_dto.product == SecurityContentProduct.splunk_app:
            conf_output = ConfOutput(input_dto.director_input_dto.input_path, input_dto.director_input_dto.config)
            conf_output.writeHeaders()
            conf_output.writeObjects(director_output_dto.detections, SecurityContentType.detections)
            conf_output.writeObjects(director_output_dto.stories, SecurityContentType.stories)
            conf_output.writeObjects(director_output_dto.baselines, SecurityContentType.baselines)
            conf_output.writeObjects(director_output_dto.investigations, SecurityContentType.investigations)
            conf_output.writeObjects(director_output_dto.lookups, SecurityContentType.lookups)
            conf_output.writeObjects(director_output_dto.macros, SecurityContentType.macros)
            conf_output.packageApp()

        elif input_dto.director_input_dto.product == SecurityContentProduct.ba_objects:
            shutil.rmtree(input_dto.output_path + '/srs/', ignore_errors=True)
            shutil.rmtree(input_dto.output_path + '/complex/', ignore_errors=True)
            os.makedirs(input_dto.output_path + '/complex/')
            os.makedirs(input_dto.output_path + '/srs/')     
            ba_yml_output = BAYmlOutput()
            ba_yml_output.writeObjects(director_output_dto.detections, input_dto.output_path)

        elif input_dto.director_input_dto.product == SecurityContentProduct.json_objects:
            api_json_output = ApiJsonOutput()
            api_json_output.writeObjects(director_output_dto.detections, input_dto.output_path, SecurityContentType.detections)
            api_json_output.writeObjects(director_output_dto.stories, input_dto.output_path, SecurityContentType.stories)
            api_json_output.writeObjects(director_output_dto.baselines, input_dto.output_path, SecurityContentType.baselines)
            api_json_output.writeObjects(director_output_dto.investigations, input_dto.output_path, SecurityContentType.investigations)
            api_json_output.writeObjects(director_output_dto.lookups, input_dto.output_path, SecurityContentType.lookups)
            api_json_output.writeObjects(director_output_dto.macros, input_dto.output_path, SecurityContentType.macros)
            api_json_output.writeObjects(director_output_dto.deployments, input_dto.output_path, SecurityContentType.deployments)

        print(f'Generate of security content successful to {conf_output.output_path}')
        return director_output_dto




class NewContentQuestions():

    @classmethod
    def get_questions_detection(self) -> list:
        questions = [
            {
                'type': 'text',
                'message': 'enter detection name',
                'name': 'detection_name',
                'default': 'Powershell Encoded Command',
            },
            {
                'type': 'text',
                'message': 'enter author name',
                'name': 'detection_author',
            },
            {
                'type': 'select',
                'message': 'select a detection type',
                'name': 'detection_type',
                'choices': [
                    'TTP',
                    'Anomaly',
                    'Hunting',
                    'Baseline',
                    'Investigation',
                    'Correlation'
                ],
                'default': 'TTP'
            },
            {
                'type': 'checkbox',
                'message': 'select the datamodels used in the detection',
                'name': 'datamodels',
                'choices': [
                    'Endpoint',
                    'Authentication',
                    'Change',
                    'Email',
                    'Network_Resolution',
                    'Network_Traffic',
                    'Network_Sessions',
                    'Updates',
                    'Vulnerabilities',
                    'Web',
                    'Risk'
                ],
                'default': 'Endpoint'
            },
            {
                'type': 'text',
                'message': 'enter search (spl)',
                'name': 'detection_search',
                'default': '| UPDATE_SPL'
            },
            {
                'type': 'text',
                'message': 'enter MITRE ATT&CK Technique IDs related to the detection, comma delimited for multiple',
                'name': 'mitre_attack_ids',
                'default': 'T1003.002'
            },
            {
                'type': 'checkbox',
                'message': 'select kill chain phases related to the detection',
                'name': 'kill_chain_phases',
                'choices': [
                    'Reconnaissance',
                    'Weaponization',
                    'Delivery',
                    'Exploitation',
                    'Installation',
                    'Command & Control',
                    'Actions on Objectives',
                    'Denial of Service'
                ],
                'default': 'Exploitation'
            },
            {
                'type': 'select',
                'message': 'security_domain for detection',
                'name': 'security_domain',
                'choices': [
                    'access',
                    'endpoint',
                    'network',
                    'threat',
                    'identity',
                    'audit'
                ],
                'default': 'endpoint'
            },
        ]
        return questions


    @classmethod
    def get_questions_story(self) -> list:
        questions = [
            {
                'type': 'text',
                'message': 'enter story name',
                'name': 'story_name',
                'default': 'Suspicious Powershell Behavior',
            },
            {
                'type': 'text',
                'message': 'enter author name',
                'name': 'story_author',
            },
            {
                'type': 'checkbox',
                'message': 'select a category',
                'name': 'category',
                'choices': [
                    'Adversary Tactics',
                    'Account Compromise',
                    'Unauthorized Software',
                    'Best Practices',
                    'Cloud Security',
                    'Command and Control',
                    'Lateral Movement',
                    'Ransomware',
                    'Privilege Escalation'
                ]
                },
                {
                    'type': 'select',
                    'message': 'select a use case',
                    'name': 'usecase',
                    'choices': [
                        'Advanced Threat Detection',
                        'Security Monitoring',
                        'Compliance',
                        'Insider Threat',
                        'Application Security',
                        'Other'
                ],
            },
        ]
        return questions


import os
import re

from contentctl.output.yml_writer import YmlWriter
from contentctl.objects.enums import SecurityContentType
from contentctl.output.finding_report_writer import FindingReportObject


class BAYmlOutput():


    def writeObjectsInPlace(self, objects: list) -> None:
        for object in objects:
            file_path = object['file_path']
            object.pop('file_path')
            object.pop('deprecated')
            object.pop('experimental') 
            YmlWriter.writeYmlFile(file_path, object)


    def writeObjects(self, objects: list, output_path: str, type: SecurityContentType = None) -> None:
        for obj in objects: 
            file_name = "ssa___" + self.convertNameToFileName(obj.name, obj.tags)
            if self.isComplexBARule(obj.search):
                file_path = os.path.join(output_path, 'complex', file_name)
            else:
                file_path = os.path.join(output_path, 'srs', file_name)

            body = FindingReportObject.writeFindingReport(obj)

            # remove unncessary fields
            YmlWriter.writeYmlFile(file_path, obj.dict(
                include =
                    {
                        "name": True,
                        "id": True,
                        "version": True,
                        "description": True,
                        "search": True,
                        "how_to_implement": True,
                        "known_false_positives": True,
                        "references": True,
                        "tags": 
                            {
                                "analytic_story": True,
                                "cis20" : True,
                                "nist": True,
                                "kill_chain_phases": True,
                                "mitre_attack_id": True,
                                "risk_severity": True,
                                "risk_score": True,
                                "security_domain": True,
                                "required_fields": True
                            },
                        "test": 
                            {
                                "name": True,
                                "tests": {
                                    '__all__': 
                                        {
                                            "name": True,
                                            "file": True,
                                            "pass_condition": True,
                                            "attack_data": {
                                                '__all__': 
                                                {
                                                    "file_name": True,
                                                    "data": True,
                                                    "source": True
                                                }
                                            }
                                        }
                                }
                            }
                    }
                ))

            # Add Finding Report Object
            with open(file_path, 'r') as file:
                data = file.read().replace('--body--', body)

            f = open(file_path, "w")
            f.write(data)
            f.close()       


    def convertNameToFileName(self, name: str, product: list):
        file_name = name \
            .replace(' ', '_') \
            .replace('-','_') \
            .replace('.','_') \
            .replace('/','_') \
            .lower()
        if 'Splunk Behavioral Analytics' in product:
            
            file_name = 'ssa___' + file_name + '.yml'
        else:
            file_name = file_name + '.yml'
        return file_name


    def isComplexBARule(self, search):
        return re.findall("stats|first_time_event|adaptive_threshold", search)







import pathlib


from pydantic import BaseModel, root_validator, validator, ValidationError, Extra, Field
from pydantic.main import ModelMetaclass
from dataclasses import dataclass
from datetime import datetime
from typing import Union

import validators

from contentctl.objects.enums import SecurityContentProduct

from contentctl.helper.utils import Utils  

from semantic_version import Version

import git
ALWAYS_PULL = True

SPLUNKBASE_URL = "https://splunkbase.splunk.com/app/{uid}/release/{release}/download"

class Manifest(BaseModel):
    #Note that many of these fields are mirrored from App

    #Some information about the developer of the app 
    author_name: str = Field(default=None, title="Enter the name of the app author")
    author_email: str = Field(default=None, title="Enter a contact email for the develop(s) of the app")
    author_company: str = Field(default=None, title="Enter the company who is developing the app")

    #uid is a numeric identifier assigned by splunkbase, so
    #homemade applications will not have this
    uid: Union[int, None] = Field(default=None, title="Unique numeric identifier assigned by Splunkbase to identify your app. You can find it in the URL of your app's landing page.  If you do not have one, leave this blank.")

    #appid is basically the internal name of you app
    appid: str = Field(default=None, title="Internal name of your app.  Note that it MUST be alphanumeric with underscores, but no spaces or other special characters")
    
    #Title is the human readable name for your application
    title: str = Field(default=None, title="Human-Readable name for your app. This can include any characters you want")

    #Self explanatory
    description: Union[str,None] = Field(default=None, title="Provide a helpful description of the app.")
    release: str = Field(default=None, title="Provide a name for the current release of the app.  This MUST follow semantic version format MAJOR.MINOR.PATCH[-tag]")

    

    @validator('author_email', always=True)
    def validate_author_email(cls, v):
        print("email is")
        print(v)
        if bool(validators.email(v)) == False:
            raise(ValueError(f"Email address {v} is invalid"))
        return v
    
    @validator('release', always=True)
    def validate_release(cls, v):
        try:    
            Version(v)
        except Exception as e:
            raise(ValueError(f"The string '{v}' is not a valid Semantic Version.  For more information on Semantic Versioning, please refer to https://semver.org/"))
        
        return v


class RepoConfig(BaseModel):

    #Needs a manifest to be able to properly generate the app
    manifest:Manifest = Field(default=None, title="Manifest Object")
    repo_path: str = Field(default='.', title="Path to the root of your app")
    repo_url: Union[str,None] = Field(default=None, title="HTTP(s) path to the repo for repo_path.  If this field is blank, it will be inferred from the repo")
    main_branch: str = Field(title="Main branch of the repo.")

    
    
    
    type: SecurityContentProduct = Field(default=SecurityContentProduct.SPLUNK_ENTERPRISE_APP, title=f"What type of product would you like to build.  Choose one of {SecurityContentProduct._member_names_}")
    skip_enrichment: bool = Field(default=True, title="Whether or not to skip the enrichment processes when validating the app.  Enrichment increases the amount of time it takes to build an app significantly because it must hit a number of Web APIs.")

    input_path: str = Field(default='.', title="Path to the root of your app")
    output_path: str = Field(default='./dist', title="Path where 'generate' will write out your raw app")
    #output_path: str = Field(default='./build', title="Path where 'build' will write out your custom app")

    #test_config: TestConfig = Field(default=TestConfig, title="Test Configuration")
    
    #@validator('manifest', always=True, pre=True)
    '''
    @root_validator(pre=True)
    def validate_manifest(cls, values):
        
        try:
            print(Manifest.parse_obj(values))
        except Exception as e:
            raise(ValueError(f"error validating manifest: {str(e)}"))
        

        return values
        print("TWO")
        #return {}
        #return Manifest.parse_obj({"email":"invalid_email@gmail.com"})
    '''
    @validator('repo_path', always=True)
    def validate_repo_path(cls,v):
        
        try:
            path = pathlib.Path(v)
        except Exception as e:
            raise(ValueError(f"Error, the provided path is is not a valid path: '{v}'"))
        
        try:    
            r = git.Repo(path)
        except Exception as e:
            raise(ValueError(f"Error, the provided path is not a valid git repo: '{path}'"))
        
        try:
            
            if ALWAYS_PULL:
                r.remotes.origin.pull()
        except Exception as e:
            raise ValueError(f"Error pulling git repository {v}: {str(e)}")
        
        
        return v


    @validator('repo_url', always=True)
    def validate_repo_url(cls, v, values):
        Utils.check_required_fields('repo_url', values, ['repo_path'])

        #First try to get the value from the repo
        try:
            remote_url_from_repo = git.Repo(values['repo_path']).remotes.origin.url
        except Exception as e:
            raise(ValueError(f"Error reading remote_url from the repo located at {values['repo_path']}"))
        
        if v is not None and remote_url_from_repo != v:
            raise(ValueError(f"The url of the remote repo supplied in the config file {v} does not "\
                              f"match the value read from the repository at {values['repo_path']}, {remote_url_from_repo}"))
        
        
        if v is None:    
            v = remote_url_from_repo

        #Ensure that the url is the proper format
        try:
            if bool(validators.url(v)) == False:
                raise(Exception)
        except:
            raise(ValueError(f"Error validating the repo_url. The url is not valid: {v}"))
        

        return v

    @validator('main_branch', always=True)
    def valid_main_branch(cls, v, values):
        Utils.check_required_fields('main_branch', values, ['repo_path', 'repo_url'])

        try:
            Utils.validate_git_branch_name(values['repo_path'],values['repo_url'], v)
        except Exception as e:
            raise ValueError(f"Error validating main_branch: {str(e)}")
        return v


'''
Initializes a Splunk Content Project
'''

from pathlib import Path
import yaml
import sys
import questionary
import os
from contentctl.objects.enums import LogLevel

import abc
from pydantic import BaseModel, Field

from contentctl.objects.config import Config

DEFAULT_FOLDERS = ['detections', 'stories', 'lookups', 'macros', 'baselines', 'dist']


def create_folders(path):

    for folder in DEFAULT_FOLDERS:
        folder_path = path + "/" + folder
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)



def NewContentPack(args, default_config):
    """
    new function creates a new configuration file based on the user input on the terminal.
    :param config: python dictionary having the configuration 
    :return: No return value
    """
    contentctl_config_file = Path(args.config)
    if contentctl_config_file.is_file():
        questions = [
            {
                'type': 'confirm',
                'message': 'File {0} already exist, are you sure you want to continue?\nTHIS WILL OVERWRITE YOUR CURRENT CONFIG!'.format(contentctl_config_file),
                'name': 'continue',
                'default': True,
            },
        ]

        answers = questionary.prompt(questions)
        if answers['continue']:
            print("> continuing with contentctl configuration...")
        else:
            print("> exiting, to create a unique configuration file in another location use the --config flag")
            sys.exit(0)


    # configuration parameters  
    if os.path.exists(args.output):  
        config_path = args.output + "/" + str(contentctl_config_file)
    else:
        print("ERROR, output folder: {0} does not exist".format(args.output))
        sys.exit(1)

    # deal with skipping configuration
    if args.skip_configuration:
        print("initializing with default configuration: {0}".format(config_path))
        # write config file
        with open(config_path, 'w') as outfile:
            yaml.dump(default_config, outfile, default_flow_style=False, sort_keys=False)

        # write folder structure
        create_folders(args.output)
        sys.exit(0)
        

    questions = [
        {
            "type": "select",
            "message": "Which build format should we use for this content pack? Builds will be created under the dist/ folder.",
            "name": "product",
            "choices": ["Splunk App", "JSON API Objects", "BA Objects", "All"],
            "default": "Splunk App"
        },
        {
            'type': 'text',
            'message': 'What should the Splunk App for this content pack be called?',
            'name': 'product_app_name',
            'default': 'Capybara Splunk Content Pack',
            'when': lambda answers: answers['product'] == "Splunk App" or answers['product'] == "All",

        },
        {
            'type': 'confirm',
            'message': 'Should this content pack be deployed to a (Cloud) Splunk Enterprise Server?',
            'name': 'deploy_to_splunk',
            'default': False,

        },
        {
            'type': 'text',
            'message': 'What is the <host>:<port> of the (Cloud) Splunk Enterprise Server?',
            'name': 'deploy_to_splunk_server',
            'default': '127.0.0.1:8089',
            'when': lambda answers: answers['deploy_to_splunk'],

        },
        {
            'type': 'text',
            'message': 'What is the username of the (Cloud) Splunk Enterprise Server?',
            'name': 'deploy_to_splunk_username',
            'default': 'admin',
            'when': lambda answers: answers['deploy_to_splunk'],

        },
        {
            'type': 'text',
            'message': 'What is the password of the (Cloud) Splunk Enterprise Server?',
            'name': 'deploy_to_splunk_password',
            'default': 'xxx',
            'when': lambda answers: answers['deploy_to_splunk'],

        },    
        {
            'type': 'text',
            'message': 'How often should analytics run? The schedule is on cron format (https://crontab.guru/).',
            'name': 'scheduling_cron_schedule',
            'default': '0 * * * *',
        },
        {
            'type': 'text',
            'message': 'What is the earliest time for analytics? Uses Splunk time modifiers (https://docs.splunk.com/Documentation/SCS/current/Search/Timemodifiers).',
            'name': 'scheduling_earliest_time',
            'default': '-70m@m',
        },
        {
            'type': 'text',
            'message': 'What is the latest time for analytics? Uses Splunk time modifiers (https://docs.splunk.com/Documentation/SCS/current/Search/Timemodifiers).',
            'name': 'scheduling_latest_time',
            'default': '-10m@m',
        },
        {
            'type': 'checkbox',
            'message': 'What should the default action be when an analytic triggers?',
            'name': 'default_actions',
            'choices': ["notable", "risk_event", "email"],
            'default': 'notable',
        },
        {
            'type': 'text',
            'message': 'What email address should we send the alerts to?',
            'name': 'to_email',
            'default': 'geralt@monsterkiller.com',
            'when': lambda answers: 'email' in answers['default_actions'],
        },
        {
            'type': 'confirm',
            'message': 'Should we include some example content? This will add a detection and its test with supporting components like lookups and macros.',
            'name': 'pre_populate',
            'default': True,
        },
    ]

    answers = questionary.prompt(questions)

    # create a custom config object to store answers
    custom_config = default_config

    # remove other product settings
    if answers['product'] == 'Splunk App':
        # pop other configs out
        custom_config['build'].pop('json_objects')
        custom_config['build'].pop('ba_objects')
        # capture configs
        custom_config['build']['splunk_app']['name'] = answers['product_app_name']
        custom_config['build']['splunk_app']['path'] = 'dist/' + answers['product_app_name'].lower().replace(" ", "_")
        custom_config['build']['splunk_app']['prefix'] = answers['product_app_name'].upper()[0: 3]

    elif answers['product'] == 'JSON API Objects':
        custom_config['build'].pop('splunk_app')
        custom_config['build'].pop('ba_objects')
    elif answers['product'] == 'BA Objects':
        custom_config['build'].pop('splunk_app')
        custom_config['build'].pop('json_objects')
    else:
        # splunk app config
        custom_config['build']['splunk_app']['name'] = answers['product_app_name']
        custom_config['build']['splunk_app']['path'] = 'dist/' + answers['product_app_name'].lower().replace(" ", "_")
        custom_config['build']['splunk_app']['prefix'] = answers['product_app_name'].upper()[0: 3]

    if answers['deploy_to_splunk']:
        custom_config['deploy']['server'] = answers['deploy_to_splunk_server']
        custom_config['deploy']['username'] = answers['deploy_to_splunk_username']
        custom_config['deploy']['password'] = answers['deploy_to_splunk_password']
    else:
        custom_config.pop('deploy')
    
    custom_config['scheduling']['cron_schedule'] = answers['scheduling_cron_schedule']
    custom_config['scheduling']['earliest_time'] = answers['scheduling_earliest_time']
    custom_config['scheduling']['latest_time'] = answers['scheduling_latest_time']

    if 'notable' in answers['default_actions']:
            custom_config['alert_actions']['notable']['rule_description'] = '%description%'
            custom_config['alert_actions']['notable']['rule_title'] = '%name%'
            custom_config['alert_actions']['notable']['nes_fields'] = ['user','dest','src']
    else:
        custom_config['alert_actions'].pop('notable')
    if 'risk_event' in answers['default_actions']:
            rba = dict()
            custom_config['alert_actions']['rba'] = rba
            custom_config['alert_actions']['rba']['enabled'] = 'true'
            
    if 'email' in answers['default_actions']:
            email = dict()
            custom_config['alert_actions']['email'] = email
            custom_config['alert_actions']['email']['subject'] = 'Alert %name% triggered'
            custom_config['alert_actions']['email']['message'] = 'The rule %name% triggered base on %description%'
            custom_config['alert_actions']['email']['to'] = answers['to_email']
        
   
    # write config file
    with open(config_path, 'w') as outfile:
        yaml.dump(custom_config, outfile, default_flow_style=False, sort_keys=False)
    print('Content pack configuration created under: {0} .. edit to fine tune details before building'.format(config_path))

    # write folder structure
    create_folders(args.output)
    print('The following folders were created: {0} under {1}.\nContent pack has been initialized, please run `new` to create new content.'.format(DEFAULT_FOLDERS, args.output))

    print("Load the custom_config into the pydantic model we have created")
    cfg = ContentPackConfig().parse_obj(custom_config)
    import pprint
    pprint.pprint(cfg.__dict__)
    print("********************")
    pprint.pprint(custom_config)
    print("done")
        
















import yaml


class YmlWriter:

    @staticmethod
    def writeYmlFile(file_path : str, obj : dict) -> None:

        with open(file_path, 'w') as outfile:
            yaml.dump(obj, outfile, default_flow_style=False, sort_keys=False)


from typing import Dict

import yaml
import sys

class YmlReader():

    @staticmethod
    def load_file(file_path: str) -> Dict:
        try:
            file_handler = open(file_path, 'r', encoding="utf-8")
            try:
                yml_obj = list(yaml.safe_load_all(file_handler))[0]
            except yaml.YAMLError as exc:
                print(exc)
                sys.exit(1)

        except OSError as exc:
            print(exc)
            sys.exit(1)

        yml_obj['file_path'] = file_path

        if 'deprecated' in file_path:
            yml_obj['deprecated'] = True
        else:
            yml_obj['deprecated'] = False

        if 'experimental' in file_path:
            yml_obj['experimental'] = True
        else:
            yml_obj['experimental'] = False

        return yml_obj



import os

from dataclasses import dataclass

from contentctl.input.director import DirectorInputDto, Director, DirectorOutputDto
from contentctl.output.svg_output import SvgOutput
from contentctl.output.attack_nav_output import AttackNavOutput


@dataclass(frozen=True)
class ReportingInputDto:
    director_input_dto: DirectorInputDto


class Reporting:

    def execute(self, input_dto: ReportingInputDto) -> None:
        director_output_dto = DirectorOutputDto([],[],[],[],[],[],[],[])
        director = Director(director_output_dto)
        director.execute(input_dto.director_input_dto)

        #svg_output = SvgOutput()
        #svg_output.writeObjects(director_output_dto.detections, input_dto.output_path)
        
        attack_nav_output = AttackNavOutput()        
        attack_nav_output.writeObjects(
            director_output_dto.detections, 
            os.path.join(input_dto.director_input_dto.input_path, "reporting")
        )
        
        print('Reporting of security content successful.')


import os
import sys

from dataclasses import dataclass
from pydantic import ValidationError



from contentctl.objects.detection import Detection
from contentctl.objects.story import Story
from contentctl.objects.baseline import Baseline
from contentctl.objects.investigation import Investigation
from contentctl.objects.playbook import Playbook
from contentctl.objects.deployment import Deployment
from contentctl.objects.macro import Macro
from contentctl.objects.lookup import Lookup
from contentctl.objects.unit_test import UnitTest

from contentctl.input.basic_builder import BasicBuilder
from contentctl.input.detection_builder import DetectionBuilder
from contentctl.input.playbook_builder import PlaybookBuilder
from contentctl.input.baseline_builder import BaselineBuilder
from contentctl.input.investigation_builder import InvestigationBuilder
from contentctl.input.story_builder import StoryBuilder
from contentctl.objects.enums import SecurityContentType
from contentctl.objects.enums import SecurityContentProduct
from contentctl.helper.utils import Utils
from contentctl.enrichments.attack_enrichment import AttackEnrichment
from contentctl.objects.config import Config

from contentctl.objects.config import Config



@dataclass(frozen=True)
class DirectorInputDto:
    input_path: str
    product: SecurityContentProduct
    config: Config


@dataclass()
class DirectorOutputDto:
     detections: list[Detection]
     stories: list[Story]
     baselines: list[Baseline]
     investigations: list[Investigation]
     playbooks: list[Playbook]
     macros: list[Macro]
     lookups: list[Lookup]
     tests: list[UnitTest]


class Director():
    input_dto: DirectorInputDto
    output_dto: DirectorOutputDto
    basic_builder: BasicBuilder
    playbook_builder: PlaybookBuilder
    baseline_builder: BaselineBuilder
    investigation_builder: InvestigationBuilder
    story_builder: StoryBuilder
    detection_builder: DetectionBuilder
    attack_enrichment: dict
    config: Config


    def __init__(self, output_dto: DirectorOutputDto) -> None:
        self.output_dto = output_dto
        self.attack_enrichment = dict()


    def execute(self, input_dto: DirectorInputDto) -> None:
        self.input_dto = input_dto
        
        
        if self.input_dto.config.enrichments.attack_enrichment:
            self.attack_enrichment = AttackEnrichment.get_attack_lookup(self.input_dto.input_path)
        
        self.basic_builder = BasicBuilder()
        self.playbook_builder = PlaybookBuilder(self.input_dto.input_path)
        self.baseline_builder = BaselineBuilder()
        self.investigation_builder = InvestigationBuilder()
        self.story_builder = StoryBuilder()
        self.detection_builder = DetectionBuilder()

        if self.input_dto.product == SecurityContentProduct.splunk_app or self.input_dto.product == SecurityContentProduct.json_objects:
            self.createSecurityContent(SecurityContentType.unit_tests)
            self.createSecurityContent(SecurityContentType.lookups)
            self.createSecurityContent(SecurityContentType.macros)
            self.createSecurityContent(SecurityContentType.baselines)
            self.createSecurityContent(SecurityContentType.investigations)
            self.createSecurityContent(SecurityContentType.playbooks)
            self.createSecurityContent(SecurityContentType.detections)
            self.createSecurityContent(SecurityContentType.stories)

        elif self.input_dto.product == SecurityContentProduct.ba_objects:
            self.createSecurityContent(SecurityContentType.unit_tests)
            self.createSecurityContent(SecurityContentType.detections)
            

    def createSecurityContent(self, type: SecurityContentType) -> None:
        objects = []
        if type == SecurityContentType.unit_tests:
            files = Utils.get_all_yml_files_from_directory(os.path.join(self.input_dto.input_path, 'tests'))
        else:
            files = Utils.get_all_yml_files_from_directory(os.path.join(self.input_dto.input_path, str(type.name)))

        validation_error_found = False
                
        already_ran = False
        progress_percent = 0

        if self.input_dto.product == SecurityContentProduct.splunk_app or self.input_dto.product == SecurityContentProduct.json_objects:
            security_content_files = [f for f in files if 'ssa___' not in f]
        elif self.input_dto.product == SecurityContentProduct.ba_objects:
            security_content_files = [f for f in files if 'ssa___' in f]
        else:
            raise(Exception(f"Cannot createSecurityContent for unknown product '{self.input_dto.product}'"))


        for index,file in enumerate(security_content_files):
            progress_percent = ((index+1)/len(security_content_files)) * 100
            try:
                type_string = "UNKNOWN TYPE"
                if type == SecurityContentType.lookups:
                        self.constructLookup(self.basic_builder, file)
                        lookup = self.basic_builder.getObject()
                        self.output_dto.lookups.append(lookup)
                
                elif type == SecurityContentType.macros:
                        self.constructMacro(self.basic_builder, file)
                        macro = self.basic_builder.getObject()
                        self.output_dto.macros.append(macro)
                
                elif type == SecurityContentType.deployments:
                        self.constructDeployment(self.basic_builder, file)
                        deployment = self.basic_builder.getObject()
                        self.output_dto.deployments.append(deployment)
                
                elif type == SecurityContentType.playbooks:
                        self.constructPlaybook(self.playbook_builder, file)
                        playbook = self.playbook_builder.getObject()
                        self.output_dto.playbooks.append(playbook)                    
                
                elif type == SecurityContentType.baselines:
                        type_string = "Baselines"
                        self.constructBaseline(self.baseline_builder, file)
                        baseline = self.baseline_builder.getObject()
                        self.output_dto.baselines.append(baseline)
                
                elif type == SecurityContentType.investigations:
                        self.constructInvestigation(self.investigation_builder, file)
                        investigation = self.investigation_builder.getObject()
                        self.output_dto.investigations.append(investigation)

                elif type == SecurityContentType.stories:
                        type_string = "Stories"
                        self.constructStory(self.story_builder, file)
                        story = self.story_builder.getObject()
                        self.output_dto.stories.append(story)
            
                elif type == SecurityContentType.detections:
                        type_string = "Detections"
                        self.constructDetection(self.detection_builder, file)
                        detection = self.detection_builder.getObject()
                        self.output_dto.detections.append(detection)
            
                elif type == SecurityContentType.unit_tests:
                        self.constructTest(self.basic_builder, file)
                        test = self.basic_builder.getObject()
                        self.output_dto.tests.append(test)

                else:
                        raise Exception(f"Unsupported type: [{type}]")
                
                if (sys.stdout.isatty() and sys.stdin.isatty() and sys.stderr.isatty()) or not already_ran:
                        already_ran = True
                        print(f"\r{f'{type_string} Progress'.rjust(23)}: [{progress_percent:3.0f}%]...", end="", flush=True)
            
            except ValidationError as e:
                print('\nValidation Error for file ' + file)
                print(e)
                validation_error_found = True

        print(f"\r{f'{type.name} Progress'.rjust(23)}: [{progress_percent:3.0f}%]...", end="", flush=True)
        print("Done!")

        if validation_error_found:
            sys.exit(1)


    def constructDetection(self, builder: DetectionBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)
        builder.addDeployment(self.input_dto.config.detection_configuration)
        builder.addRBA()
        builder.addProvidingTechnologies()
        builder.addNesFields()
        builder.addAnnotations()
        builder.addMappings()
        builder.addBaseline(self.output_dto.baselines)
        builder.addPlaybook(self.output_dto.playbooks)
        builder.addUnitTest(self.output_dto.tests)
        builder.addMacros(self.output_dto.macros)
        builder.addLookups(self.output_dto.lookups)
        
        if self.input_dto.config.enrichments.attack_enrichment:
            builder.addMitreAttackEnrichment(self.attack_enrichment)

        if self.input_dto.config.enrichments.cve_enrichment:
            builder.addCve()
    
        if self.input_dto.config.enrichments.splunk_app_enrichment:
            builder.addSplunkApp()


    def constructStory(self, builder: StoryBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)
        builder.addDetections(self.output_dto.detections, self.input_dto.config)
        builder.addInvestigations(self.output_dto.investigations)
        builder.addBaselines(self.output_dto.baselines)
        builder.addAuthorCompanyName()


    def constructBaseline(self, builder: BaselineBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)
        builder.addDeployment(self.output_dto.deployments)


    def constructDeployment(self, builder: BasicBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path, SecurityContentType.deployments)

    def constructLookup(self, builder: BasicBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path, SecurityContentType.lookups)


    def constructMacro(self, builder: BasicBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path, SecurityContentType.macros)


    def constructPlaybook(self, builder: PlaybookBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)
        builder.addDetections()


    def constructTest(self, builder: BasicBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path, SecurityContentType.unit_tests)


    def constructInvestigation(self, builder: InvestigationBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)
        builder.addInputs()
        builder.addLowercaseName()

    def constructObjects(self, builder: BasicBuilder, file_path: str) -> None:
        builder.reset()
        builder.setObject(file_path)


import sys
import re
import os

from pydantic import ValidationError

from contentctl.input.yml_reader import YmlReader
from contentctl.objects.detection import Detection
from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.macro import Macro
from contentctl.objects.mitre_attack_enrichment import MitreAttackEnrichment
from contentctl.enrichments.cve_enrichment import CveEnrichment
from contentctl.enrichments.splunk_app_enrichment import SplunkAppEnrichment
from contentctl.objects.config import ConfigDetectionConfiguration


class DetectionBuilder():
    security_content_obj : SecurityContentObject


    def setObject(self, path: str) -> None:
        yml_dict = YmlReader.load_file(path)
        yml_dict["tags"]["name"] = yml_dict["name"]
        self.security_content_obj = Detection.parse_obj(yml_dict)
        self.security_content_obj.source = os.path.split(os.path.dirname(self.security_content_obj.file_path))[-1]      


    def addDeployment(self, detection_configuration: ConfigDetectionConfiguration) -> None:
        if self.security_content_obj:
            self.security_content_obj.deployment = detection_configuration


    def addRBA(self) -> None:
        if self.security_content_obj:
            risk_objects = []
            risk_object_user_types = {'user', 'username', 'email address'}
            risk_object_system_types = {'device', 'endpoint', 'hostname', 'ip address'}

            if hasattr(self.security_content_obj.tags, 'observable') and hasattr(self.security_content_obj.tags, 'risk_score'):
                for entity in self.security_content_obj.tags.observable:
                    risk_object = dict()
                    if entity['type'].lower() in risk_object_user_types:
                        for r in entity['role']:
                            if 'attacker' == r.lower() or 'victim' ==r.lower():
                                risk_object['risk_object_type'] = 'user'
                                risk_object['risk_object_field'] = entity['name']
                                risk_object['risk_score'] = self.security_content_obj.tags.risk_score
                                risk_objects.append(risk_object)

                    elif entity['type'].lower() in risk_object_system_types:
                        for r in entity['role']:
                            if 'attacker' == r.lower() or 'victim' ==r.lower():
                                risk_object['risk_object_type'] = 'system'
                                risk_object['risk_object_field'] = entity['name']
                                risk_object['risk_score'] = self.security_content_obj.tags.risk_score
                                risk_objects.append(risk_object)
                    else:
                        risk_object['threat_object_field'] = entity['name']
                        risk_object['threat_object_type'] = entity['type'].lower()
                        risk_objects.append(risk_object)
                        continue

            if self.security_content_obj.tags.risk_score >= 80:
                self.security_content_obj.tags.risk_severity = 'high'
            elif (self.security_content_obj.tags.risk_score >= 50 and self.security_content_obj.tags.risk_score <= 79):
                self.security_content_obj.tags.risk_severity = 'medium'
            else:
                self.security_content_obj.tags.risk_severity = 'low'

            self.security_content_obj.risk = risk_objects


    def addProvidingTechnologies(self) -> None:
        if self.security_content_obj:
            # if self.security_content_obj.tags.supported_tas:
                if 'Endpoint' in self.security_content_obj.datamodel:
                    self.security_content_obj.providing_technologies = ["Sysmon", "Microsoft Windows","Carbon Black Response","CrowdStrike Falcon", "Symantec Endpoint Protection"]
                if "`cloudtrail`" in str(self.security_content_obj.search):
                    self.security_content_obj.providing_technologies = ["Amazon Web Services - Cloudtrail"]
                if '`wineventlog_security`' in self.security_content_obj.search or '`powershell`' in self.security_content_obj.search:
                    self.security_content_obj.providing_technologies = ["Microsoft Windows"]

    
    def addNesFields(self) -> None:
        if self.security_content_obj:
            if self.security_content_obj.deployment:
                if self.security_content_obj.deployment.notable:
                    nes_fields = ",".join(list(self.security_content_obj.deployment.notable.nes_fields))
                    self.security_content_obj.nes_fields = nes_fields
                    

    def addMappings(self) -> None:
        if self.security_content_obj:
            keys = ['mitre_attack', 'kill_chain_phases', 'cis20', 'nist']
            mappings = {}
            for key in keys:
                if key == 'mitre_attack':
                    if getattr(self.security_content_obj.tags, 'mitre_attack_id'):
                        mappings[key] = getattr(self.security_content_obj.tags, 'mitre_attack_id')
                elif getattr(self.security_content_obj.tags, key):
                    mappings[key] = getattr(self.security_content_obj.tags, key)
            self.security_content_obj.mappings = mappings


    def addAnnotations(self) -> None:
        if self.security_content_obj:
            annotations = {}
            annotation_keys = ['mitre_attack', 'kill_chain_phases', 'cis20', 'nist', 
                'analytic_story', 'observable', 'context', 'impact', 'confidence', 'cve']
            for key in annotation_keys:
                if key == 'mitre_attack':
                    if getattr(self.security_content_obj.tags, 'mitre_attack_id'):
                        annotations[key] = getattr(self.security_content_obj.tags, 'mitre_attack_id')
                try:
                    if getattr(self.security_content_obj.tags, key):
                        annotations[key] = getattr(self.security_content_obj.tags, key)
                except AttributeError as e:
                    continue
            self.security_content_obj.annotations = annotations    


    def addPlaybook(self, playbooks: list) -> None:
        if self.security_content_obj:
            matched_playbooks = []
            for playbook in playbooks:
                if playbook.tags.detections:
                    for detection in playbook.tags.detections:
                        if detection == self.security_content_obj.name:
                            matched_playbooks.append(playbook)

            self.security_content_obj.playbooks = matched_playbooks


    def addBaseline(self, baselines: list) -> None:
        if self.security_content_obj:
            matched_baselines = []
            for baseline in baselines:
                for detection in baseline.tags.detections:
                    if detection == self.security_content_obj.name:
                        matched_baselines.append(baseline)

            self.security_content_obj.baselines = matched_baselines


    def addUnitTest(self, tests: list) -> None:
        if self.security_content_obj:
            for test in tests:
                if test.name == f"{self.security_content_obj.name} Unit Test":
                    self.security_content_obj.test = test
                    return
            if self.security_content_obj.type not in ["Correlation"] and \
               self.security_content_obj.deprecated == False and \
               self.security_content_obj.experimental == False:
                #raise(Exception(f"No tests found found {self.security_content_obj.file_path}"))
                print(f"No tests found found {self.security_content_obj.file_path}")
            return None



    def addMitreAttackEnrichment(self, attack_enrichment: dict) -> None:
        if self.security_content_obj:
            if attack_enrichment:
                if self.security_content_obj.tags.mitre_attack_id:
                    self.security_content_obj.tags.mitre_attack_enrichments = []
                    for mitre_attack_id in self.security_content_obj.tags.mitre_attack_id:
                        if mitre_attack_id in attack_enrichment:
                            mitre_attack_enrichment = MitreAttackEnrichment(
                                mitre_attack_id = mitre_attack_id, 
                                mitre_attack_technique = attack_enrichment[mitre_attack_id]["technique"], 
                                mitre_attack_tactics = sorted(attack_enrichment[mitre_attack_id]["tactics"]), 
                                mitre_attack_groups = sorted(attack_enrichment[mitre_attack_id]["groups"])
                            )
                            self.security_content_obj.tags.mitre_attack_enrichments.append(mitre_attack_enrichment)
                        else:
                            #print("mitre_attack_id " + mitre_attack_id + " doesn't exist for detecction " + self.security_content_obj.name)
                            raise ValueError("mitre_attack_id " + mitre_attack_id + " doesn't exist for detection " + self.security_content_obj.name)


    def addMacros(self, macros: list) -> None:
        if self.security_content_obj:
            macros_found = re.findall(r'`([^\s]+)`', self.security_content_obj.search)
            macros_filtered = set()
            self.security_content_obj.macros = []

            for macro in macros_found:
                if not '_filter' in macro and not 'drop_dm_object_name' in macro:
                    start = macro.find('(')
                    if start != -1:
                        macros_filtered.add(macro[:start])
                    else:
                        macros_filtered.add(macro)

            for macro_name in macros_filtered:
                for macro in macros:
                    if macro_name == macro.name:
                        self.security_content_obj.macros.append(macro)

            name = self.security_content_obj.name.replace(' ', '_').replace('-', '_').replace('.', '_').replace('/', '_').lower() + '_filter'
            macro = Macro(name=name, definition='search *', description='Update this macro to limit the output results to filter out false positives.')
            
            self.security_content_obj.macros.append(macro)


    def addLookups(self, lookups: list) -> None:
        if self.security_content_obj:
            lookups_found = re.findall(r'lookup (?:update=true)?(?:append=t)?\s*([^\s]*)', self.security_content_obj.search)
            self.security_content_obj.lookups = []
            for lookup_name in lookups_found:
                for lookup in lookups:
                    if lookup.name == lookup_name:
                        self.security_content_obj.lookups.append(lookup)


    def addCve(self) -> None:
        if self.security_content_obj:
            self.security_content_obj.cve_enrichment = []
            if self.security_content_obj.tags.cve:
                for cve in self.security_content_obj.tags.cve:
                    self.security_content_obj.cve_enrichment.append(CveEnrichment.enrich_cve(cve))

    def addSplunkApp(self) -> None:
        if self.security_content_obj:
            self.security_content_obj.splunk_app_enrichment = []
            if self.security_content_obj.tags.supported_tas:
                for splunk_app in self.security_content_obj.tags.supported_tas:
                    self.security_content_obj.splunk_app_enrichment.append(SplunkAppEnrichment.enrich_splunk_app(splunk_app))

    def reset(self) -> None:
        self.security_content_obj = None


    def getObject(self) -> SecurityContentObject:
        return self.security_content_obj



import os


from contentctl.objects.enums import SecurityContentType
from contentctl.output.attack_nav_writer import AttackNavWriter


class AttackNavOutput():

    def writeObjects(self, objects: list, output_path: str, type: SecurityContentType = None) -> None:
        techniques = dict()
        for detection in objects:
            if detection.tags.mitre_attack_enrichments:
                for mitre_attack_enrichment in detection.tags.mitre_attack_enrichments:
                    if not mitre_attack_enrichment.mitre_attack_id in techniques:
                        techniques[mitre_attack_enrichment.mitre_attack_id] = {
                                'score': 1,
                                'file_paths': ['https://github.com/splunk/security_content/blob/develop/detections/' + detection.source + '/' + self.convertNameToFileName(detection.name)]
                            }
                    else:
                        techniques[mitre_attack_enrichment.mitre_attack_id]['score'] = techniques[mitre_attack_enrichment.mitre_attack_id]['score'] + 1
                        techniques[mitre_attack_enrichment.mitre_attack_id]['file_paths'].append('https://github.com/splunk/security_content/blob/develop/detections/' + detection.source + '/' + self.convertNameToFileName(detection.name))

        AttackNavWriter.writeAttackNavFile(techniques, os.path.join(output_path, 'coverage.json'))


    def convertNameToFileName(self, name: str):
        file_name = name \
            .replace(' ', '_') \
            .replace('-','_') \
            .replace('.','_') \
            .replace('/','_') \
            .lower()
        file_name = file_name + '.yml'
        return file_name



import os

from dataclasses import dataclass

from contentctl.input.director import DirectorInputDto, Director, DirectorOutputDto
from contentctl.output.doc_md_output import DocMdOutput


@dataclass(frozen=True)
class DocGenInputDto:
    director_input_dto: DirectorInputDto

class DocGen:

    def execute(self, input_dto: DocGenInputDto) -> None:
        director_output_dto = DirectorOutputDto([],[],[],[],[],[],[],[])
        director = Director(director_output_dto)
        director.execute(input_dto.director_input_dto)

        doc_md_output = DocMdOutput()
        doc_md_output.writeObjects(
            [director_output_dto.stories, director_output_dto.detections, director_output_dto.playbooks], 
            os.path.join(input_dto.director_input_dto.input_path, "docs")
        )

        print('Generating Docs of security content successful.')


from dataclasses import dataclass

from contentctl.objects.test_config import TestConfig

from contentctl.input.director import DirectorOutputDto
from contentctl.actions.detection_testing.GitHubService import (
    GithubService,
)

from contentctl.actions.detection_testing.DetectionTestingManager import (
    DetectionTestingManager,
    DetectionTestingManagerInputDto,
)


from contentctl.actions.detection_testing.infrastructures.DetectionTestingInfrastructure import (
    DetectionTestingManagerOutputDto,
)


from contentctl.actions.detection_testing.views.DetectionTestingViewWeb import (
    DetectionTestingViewWeb,
)

from contentctl.actions.detection_testing.views.DetectionTestingViewCLI import (
    DetectionTestingViewCLI,
)

from contentctl.actions.detection_testing.views.DetectionTestingViewFile import (
    DetectionTestingViewFile,
)

from argparse import Namespace


MAXIMUM_CONFIGURATION_TIME_SECONDS = 600


@dataclass(frozen=True)
class TestInputDto:
    director_output_dto: DirectorOutputDto
    githubService: GithubService
    config: TestConfig


class TestOutputDto:
    results: list


class Test:
    def execute(self, input_dto: TestInputDto) -> bool:

        test_director = input_dto.githubService.get_all_content(
            input_dto.director_output_dto
        )

        output_dto = DetectionTestingManagerOutputDto()

        web = DetectionTestingViewWeb(config=input_dto.config, sync_obj=output_dto)
        cli = DetectionTestingViewCLI(config=input_dto.config, sync_obj=output_dto)
        file = DetectionTestingViewFile(config=input_dto.config, sync_obj=output_dto)

        manager_input_dto = DetectionTestingManagerInputDto(
            config=input_dto.config,
            testContent=test_director,
            views=[web, cli, file],
        )
        manager = DetectionTestingManager(
            input_dto=manager_input_dto, output_dto=output_dto
        )

        manager.setup()
        manager.execute()

        try:
            summary_results = file.getSummaryObject()
            summary = summary_results.get("summary", {})

            print("Test Summary")
            print(f"\tSuccess                      : {summary.get('success',False)}")
            print(
                f"\tSuccess Rate                 : {summary.get('success_rate','ERROR')}"
            )
            print(
                f"\tTotal Detections             : {summary.get('total_detections','ERROR')}"
            )
            print(
                f"\tPassed Detections            : {summary.get('total_pass','ERROR')}"
            )
            print(
                f"\tFailed or Untested Detections: {summary.get('total_fail_or_untested','ERROR')}"
            )
            print(f"\tTest Results File            : {file.getOutputFilePath()}")
            return summary_results.get("summary", {}).get("success", False)

        except Exception as e:
            print(f"Error determining if whole test was successful: {str(e)}")
            return False





from pydantic import BaseModel, validator, ValidationError

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.unit_test_test import UnitTestTest

class UnitTest(BaseModel, SecurityContentObject):
    name: str
    tests: list[UnitTestTest]
    file_path: str
    


from pydantic import BaseModel, validator, ValidationError


class ConfigGlobal(BaseModel):
    log_path: str
    log_level: str


class ConfigScheduling(BaseModel):
    cron_schedule: str
    earliest_time: str
    latest_time: str
    schedule_window: str


class ConfigNotable(BaseModel):
    rule_description: str
    rule_title: str
    nes_fields: list


class ConfigEmail(BaseModel):
    subject: str
    to: str
    message: str


class ConfigSlack(BaseModel):
    channel: str
    message: str


class ConfigPhantom(BaseModel):
    cam_workers: str
    label: str
    phantom_server: str
    sensitivity: str
    severity: str


class ConfigRba(BaseModel):
    enabled: str


class ConfigDetectionConfiguration(BaseModel):
    scheduling: ConfigScheduling
    notable: ConfigNotable = None
    email: ConfigEmail = None
    slack: ConfigSlack = None
    phantom: ConfigPhantom = None
    rba: ConfigRba = None


class ConfigAlertAction(BaseModel):
    notable: ConfigNotable


class ConfigTest(BaseModel):
    docker_image: str
    apps: list
    assets: str


class ConfigDeploy(BaseModel):
    app: str
    username: str
    password: str
    server: str


class ConfigBuildSplunk(BaseModel):
    path: str
    name: str
    prefix: str
    author: str
    author_email: str


class ConfigBuildJson(BaseModel):
    path: str


class ConfigBuildBa(BaseModel):
    path: str


class ConfigBuild(BaseModel):
    splunk_app: ConfigBuildSplunk
    #json_objects: ConfigBuildJson
    #ba_objects: ConfigBuildBa


class ConfigEnrichments(BaseModel):
    attack_enrichment: bool
    cve_enrichment: bool
    splunk_app_enrichment: bool



class Config(BaseModel):
    #general: ConfigGlobal
    detection_configuration: ConfigDetectionConfiguration
    test: ConfigTest = None
    deploy: ConfigDeploy
    build: ConfigBuild
    enrichments: ConfigEnrichments




import string
import uuid
import requests

from pydantic import BaseModel, validator, ValidationError
from dataclasses import dataclass
from datetime import datetime

from contentctl.objects.security_content_object import SecurityContentObject
from contentctl.objects.enums import DataModel
from contentctl.objects.baseline_tags import BaselineTags
from contentctl.objects.deployment import Deployment


class Baseline(BaseModel, SecurityContentObject):
    # baseline spec
    name: str
    id: str
    version: int
    date: str
    author: str
    type: str
    datamodel: list
    description: str
    search: str
    how_to_implement: str
    known_false_positives: str
    check_references: bool = False #Validation is done in order, this field must be defined first
    references: list
    tags: BaselineTags

    # enrichment
    deployment: Deployment = None


    @validator('name')
    def name_max_length(cls, v):
        if len(v) > 67:
            raise ValueError('name is longer then 67 chars: ' + v)
        return v

    @validator('name')
    def name_invalid_chars(cls, v):
        invalidChars = set(string.punctuation.replace("-", ""))
        if any(char in invalidChars for char in v):
            raise ValueError('invalid chars used in name: ' + v)
        return v

    @validator('id')
    def id_check(cls, v, values):
        try:
            uuid.UUID(str(v))
        except:
            raise ValueError('uuid is not valid: ' + values["name"])
        return v

    @validator('date')
    def date_valid(cls, v, values):
        try:
            datetime.strptime(v, "%Y-%m-%d")
        except:
            raise ValueError('date is not in format YYYY-MM-DD: ' + values["name"])
        return v

    @validator('type')
    def type_valid(cls, v, values):
        if v != "Baseline":
            raise ValueError('not valid analytics type: ' + values["name"])
        return v

    @validator('datamodel')
    def datamodel_valid(cls, v, values):
        for datamodel in v:
            if datamodel not in [el.name for el in DataModel]:
                raise ValueError('not valid data model: ' + values["name"])
        return v

    @validator('description', 'how_to_implement')
    def encode_error(cls, v, values, field):
        try:
            v.encode('ascii')
        except UnicodeEncodeError:
            raise ValueError('encoding error in ' + field.name + ': ' + values["name"])
        return v

    @validator('search')
    def search_validate(cls, v, values):
        # write search validator
        return v



import os
import git
import shutil
import requests
import random
import string
from timeit import default_timer
import pathlib
import datetime
from typing import Union
import tqdm
from contentctl.objects.security_content_object import SecurityContentObject
from math import ceil

TOTAL_BYTES = 0
ALWAYS_PULL = True


class Utils:
    @staticmethod
    def get_all_yml_files_from_directory(path: str) -> list:
        listOfFiles = list()
        for (dirpath, dirnames, filenames) in os.walk(path):
            for file in filenames:
                if file.endswith(".yml"):
                    listOfFiles.append(os.path.join(dirpath, file))

        return sorted(listOfFiles)

    @staticmethod
    def add_id(
        id_dict: dict[str, list[str]], obj: SecurityContentObject, path: str
    ) -> None:
        if hasattr(obj, "id"):
            obj_id = obj.id
            if obj_id in id_dict:
                id_dict[obj_id].append(path)
            else:
                id_dict[obj_id] = [path]

    # Otherwise, no ID so nothing to add....

    @staticmethod
    def check_ids_for_duplicates(id_dict: dict[str, list[str]]) -> bool:
        validation_error = False
        for key, values in id_dict.items():
            if len(values) > 1:
                validation_error = True
                id_conflicts_string = "\n\t* ".join(values)
                print(
                    f"\nError validating id [{key}] - duplicate ID is used for the following content: \n\t* {id_conflicts_string}"
                )
        return validation_error

    @staticmethod
    def validate_git_hash(
        repo_path: str, repo_url: str, commit_hash: str, branch_name: Union[str, None]
    ) -> bool:

        # Get a list of all branches
        repo = git.Repo(repo_path)
        if commit_hash is None:
            # No need to validate the hash, it was not supplied
            return True

        try:
            all_branches_containing_hash = repo.git.branch(
                "--contains", commit_hash
            ).split("\n")
            # this is a list of all branches that contain the hash.  They are in the format:
            # * <some number of spaces> branchname (if the branch contains the hash)
            # <some number of spaces>   branchname (if the branch does not contain the hash)
            # Note, of course, that a hash can be in 0, 1, more branches!
            for branch_string in all_branches_containing_hash:
                if branch_string.split(" ")[0] == "*" and (
                    branch_string.split(" ")[-1] == branch_name or branch_name == None
                ):
                    # Yes, the hash exists in the branch (or branch_name was None and it existed in at least one branch)!
                    return True
            # If we get here, it does not exist in the given branch
            raise (Exception("Does not exist in branch"))

        except Exception as e:
            if branch_name is None:
                branch_name = "ANY_BRANCH"
            if ALWAYS_PULL:
                raise (
                    ValueError(
                        f"hash '{commit_hash}' not found in '{branch_name}' for repo located at:\n  * repo_path: {repo_path}\n  * repo_url: {repo_url}"
                    )
                )
            else:
                raise (
                    ValueError(
                        f"hash '{commit_hash}' not found in '{branch_name}' for repo located at:\n  * repo_path: {repo_path}\n  * repo_url: {repo_url}"
                        "If the hash is new, try pulling the repo."
                    )
                )

    @staticmethod
    def get_default_branch_name(repo_path: str, repo_url: str) -> str:
        # Even though the default branch is only a notion in GitHub or
        # similar systems, we will consinder the default branch
        # to be the name of the branch that is the HEAD of the repo.
        # This means that it should work for ANY repo with a remote.

        repo = git.Repo(repo_path)

        # Only works for remotes!
        for remote in repo.remotes.origin.refs:
            if remote.name.endswith("/HEAD"):
                # return the name of this branch.  it will be prefixed with 'origin/', so remove the origin/
                return remote.ref.name.replace("origin/", "")
        raise (
            ValueError(
                f"Failed to find default branch in repo_path: {repo_path}\n  * repo_url: {repo_url}"
            )
        )

    @staticmethod
    def validate_git_branch_name(repo_path: str, repo_url: str, name: str) -> bool:
        # Get a list of all branches
        repo = git.Repo(repo_path)

        all_branches = [branch.name for branch in repo.refs]
        # remove "origin/" from the beginning of each branch name
        all_branches = [branch.replace("origin/", "") for branch in all_branches]

        if name in all_branches:
            return True

        else:
            if ALWAYS_PULL:
                raise (
                    ValueError(
                        f"branch '{name}' not found in repo located at:\n  * repo_path: {repo_path}\n  * repo_url: {repo_url}"
                    )
                )
            else:
                raise (
                    ValueError(
                        f"branch '{name}' not found in repo located at:\n  * repo_path: {repo_path}\n  * repo_url: {repo_url}"
                        "If the branch is new, try pulling the repo."
                    )
                )

    @staticmethod
    def validate_git_pull_request(repo_path: str, pr_number: int) -> str:
        # Get a list of all branches
        repo = git.Repo(repo_path)
        # List of all remotes that match this format.  If the PR exists, we
        # should find exactly one in the format SHA_HASH\tpull/pr_number/head
        pr_and_hash = repo.git.ls_remote("origin", f"pull/{pr_number}/head")

        if len(pr_and_hash) == 0:
            raise (
                ValueError(
                    f"pr_number {pr_number} not found in Remote '{repo.remote().url}'"
                )
            )

        pr_and_hash_lines = pr_and_hash.split("\n")
        if len(pr_and_hash_lines) > 1:
            raise (
                ValueError(
                    f"Somehow, more than 1 PR was found with pr_number {pr_number}:\n{pr_and_hash}\nThis should not happen."
                )
            )

        if pr_and_hash_lines[0].count("\t") == 1:
            hash, _ = pr_and_hash_lines[0].split("\t")
            return hash
        else:
            raise (
                ValueError(
                    f"Expected PR Format:\nCOMMIT_HASH\tpull/{pr_number}/head\nbut got\n{pr_and_hash_lines[0]}"
                )
            )

        return hash

    @staticmethod
    def check_required_fields(
        thisField: str, definedFields: dict, requiredFields: list[str]
    ):
        missing_fields = [
            field for field in requiredFields if field not in definedFields
        ]
        if len(missing_fields) > 0:
            raise (
                ValueError(
                    f"Could not validate - please resolve other errors resulting in missing fields {missing_fields}"
                )
            )

    @staticmethod
    def verify_file_exists(
        file_path: str, verbose_print=False, timeout_seconds: int = 10
    ) -> None:

        try:
            if pathlib.Path(file_path).is_file():
                # This is a file and we know it exists
                return None
        except Exception as e:
            print(f"Could not copy local file {file_path} the file because {str(e)}")

        # Try to make a head request to verify existence of the file
        try:
            req = requests.head(
                file_path, timeout=timeout_seconds, verify=True, allow_redirects=True
            )
            if req.status_code > 400:
                raise (Exception(f"Return code {req.status_code}"))
        except Exception as e:
            raise (
                Exception(
                    f"Cannot confirm the existence of '{file_path}' - are you sure it exists: {str(e)}"
                )
            )

    @staticmethod
    def copy_local_file(
        file_path: str,
        destination_file: str,
        overwrite_file: bool = True,
        verbose_print: bool = False,
    ):
        sourcePath = pathlib.Path(file_path)
        destPath = pathlib.Path(destination_file)
        if verbose_print:
            print(
                f"Copying [{sourcePath}] to [{destPath}]...",
                end="",
                flush=True,
            )
        try:
            # generates an exception only if the copyfile fails.
            # if we try this with a URL, it won't be found as a file and no
            # exception will be generated
            if sourcePath.is_file():
                if destPath.is_file():
                    if overwrite_file and not sourcePath.samefile(destPath):
                        shutil.copyfile(sourcePath, destPath)
                    else:
                        # Don't do anything, don't overwrite the file
                        pass
                elif destPath.exists():
                    raise (
                        Exception(
                            f"[{destPath}] exists, but it is not a file.  It cannot be overwritten."
                        )
                    )
                else:
                    shutil.copyfile(sourcePath, destPath)
            else:
                raise (Exception(f"[{sourcePath}] does not exist"))
        except Exception as e:
            raise (
                Exception(
                    f"Error: Could not copy local file [{sourcePath}] to [{destPath}]: [{str(e)}]"
                )
            )
        if verbose_print:
            print("Done")

    @staticmethod
    def download_file_from_http(
        file_path: str,
        destination_file: str,
        input_pbar: Union[tqdm.tqdm, None] = None,
        overwrite_file: bool = False,
    ):
        global TOTAL_BYTES
        sourcePath = pathlib.Path(file_path)
        destinationPath = pathlib.Path(destination_file)

        if input_pbar is None:
            pbar = tqdm.tqdm(
                total=100,
                desc="Downloading File",
                unit="B",
                unit_scale=True,
                bar_format=f"Downloading {sourcePath.name}".ljust(80)
                + "{percentage:3.0f}%[{bar:20}]"
                + "[{n_fmt}/{total_fmt} | ETA: {remaining}]",
            )
        else:
            pbar = input_pbar
            pbar.bar_format = (
                f"Downloading {sourcePath.name}".ljust(80)
                + "{percentage:3.0f}%[{bar:20}]"
                + "[{n_fmt}/{total_fmt} | ETA: {remaining}]"
            )
            pbar.unit = "B"
            pbar.unit_scale = True

            pbar.reset()

        if destinationPath.is_file() and overwrite_file is False:
            pbar.bar_format = (
                f"Downloading {sourcePath.name}".ljust(80)
                + "{percentage:3.0f}%[{bar:20}]"
                + "[PREVIOUSLY CACHED]"
            )
            pbar.update(100)
            if input_pbar is None:
                pbar.close()
            return
        elif destinationPath.is_file() and overwrite_file is True:
            # Overwrite the file
            pass
        elif destinationPath.exists():
            # The path exists but it is not a file.  This is an issue
            # regardless of whether or not overwrite_file is enabled

            raise (
                Exception(
                    f"[{destinationPath}] already exists, but it is not a file. We cannot overwrite it."
                )
            )

        try:
            download_start_time = default_timer()
            bytes_written = 0
            file_to_download = requests.get(file_path, stream=True)
            file_to_download.raise_for_status()
            content_length = int(file_to_download.headers["Content-length"])

            chunk_size = max(1024 * 1024, ceil(content_length / 100))
            pbar.total = content_length
            pbar.reset()

            pbar.bar_format = (
                f"Downloading {sourcePath.name}".ljust(80)
                + "{percentage:3.0f}%[{bar:20}]"
                + "[{n_fmt}/{total_fmt} | ETA: {remaining}]"
            )

            # pbar.update()

            with destinationPath.open("wb") as output:
                for piece in file_to_download.iter_content(chunk_size=chunk_size):
                    bytes_written += output.write(piece)
                    pbar.update(len(piece))

            TOTAL_BYTES += bytes_written

        except requests.exceptions.ConnectionError as e:
            raise (
                Exception(
                    f"Error: Could not download file [{file_path}] to [{destinationPath}] (Unable to connect to server. Are you sure the server exists and you have connectivity to it?): [{str(e)}]"
                )
            )

        except requests.exceptions.HTTPError as e:
            raise (
                Exception(
                    f"Error: Could not download file [{file_path}] to [{destinationPath}] (The file was probably not found on the server): [{str(e)}]"
                )
            )
        except requests.exceptions.Timeout as e:
            raise (
                Exception(
                    f"Error: Could not download file [{file_path}] to [{destinationPath}] (Timeout getting file): [{str(e)}]"
                )
            )
        except Exception as e:
            raise (
                Exception(
                    f"Error: Could not download file [{file_path}] to [{destinationPath}] (Unknown Reason): [{str(e)}]"
                )
            )
        finally:
            if input_pbar is None:
                pbar.close()
            # Otherwise, don't close it because we will keep using it

        pbar.reset()
        return

    # taken from attack_range
    @staticmethod
    def get_random_password(
        password_min_length: int = 16, password_max_length: int = 26
    ) -> str:
        random_source = string.ascii_letters + string.digits
        password = random.choice(string.ascii_lowercase)
        password += random.choice(string.ascii_uppercase)
        password += random.choice(string.digits)

        for i in range(random.randrange(password_min_length, password_max_length)):
            password += random.choice(random_source)

        password_list = list(password)
        random.SystemRandom().shuffle(password_list)
        password = "".join(password_list)
        return password

    @staticmethod
    def warning_print(
        msg: str, prefix: str = "MESSAGE TO CONTENTCTL DEV", suppress=False
    ):
        if not suppress:
            print(f"{prefix}: {msg}")

    @staticmethod
    def getFixedWidth(num: float, decimal_places: int) -> str:
        fstring = "{:." + str(decimal_places) + "f}"
        return fstring.format(num)

    @staticmethod
    def getPercent(numerator: float, denominator: float, decimal_places: int) -> str:
        if denominator == 0:
            return "UKNOWN"
        ratio = numerator / denominator
        percent = ratio * 100
        return Utils.getFixedWidth(percent, decimal_places) + "%"




import sys
import os

from pydantic import ValidationError
from pathlib import Path

from contentctl.objects.playbook import Playbook
from contentctl.input.yml_reader import YmlReader


class PlaybookBuilder():
    playbook: Playbook
    input_path: str
    
    
    def __init__(self, input_path: str):
        self.input_path = input_path

    def setObject(self, path: str) -> None:
        yml_dict = YmlReader.load_file(path)

        try:
            self.playbook = Playbook.parse_obj(yml_dict)

        except ValidationError as e:
            print('Validation Error for file ' + path)
            print(e)
            sys.exit(1)


    def addDetections(self) -> None:
        if self.playbook.tags.detections:
            self.playbook.tags.detection_objects = []
            for detection in self.playbook.tags.detections:
                detection_object = {
                    "name": detection,
                    "lowercase_name": self.convertNameToFileName(detection),
                    "path": self.findDetectionPath(detection)
                }
                self.playbook.tags.detection_objects.append(detection_object)


    def reset(self) -> None:
        self.playbook = None


    def getObject(self) -> Playbook:
        return self.playbook


    def convertNameToFileName(self, name: str):
        file_name = name \
            .replace(' ', '_') \
            .replace('-','_') \
            .replace('.','_') \
            .replace('/','_') \
            .lower()
        return file_name


    def findDetectionPath(self, detection_name: str) -> str:
        for path in Path(os.path.join(self.input_path, 'detections')).rglob(self.convertNameToFileName(detection_name) + '.yml'):
            normalized_path = os.path.normpath(path)
            path_components = normalized_path.split(os.sep)
            value_index = path_components.index('detections')
            return "/".join(path_components[value_index:])


import os
import re
from jinja2 import Environment, FileSystemLoader

from contentctl.objects.detection import Detection
from contentctl.objects.constants import *

class FindingReportObject():

    @staticmethod
    def writeFindingReport(detection : Detection) -> None:
    
        if detection.tags.confidence < 33:
            detection.tags.confidence_id = 1
        elif detection.tags.confidence < 66:
            detection.tags.confidence_id = 2
        else:
            detection.tags.confidence_id = 3

        detection.tags.context_ids = list()
        for context in detection.tags.context:
            detection.tags.context_ids.append(SES_CONTEXT_MAPPING[context])
        
        if detection.tags.impact < 20:
            detection.tags.impact_id = 1
        elif detection.tags.impact < 40:
            detection.tags.impact_id = 2
        elif detection.tags.impact < 60:
            detection.tags.impact_id = 3
        elif detection.tags.impact < 80:
            detection.tags.impact_id = 4
        else:
            detection.tags.impact_id = 5                 

        detection.tags.kill_chain_phases_id = list()
        for kill_chain_phase in detection.tags.kill_chain_phases:
            detection.tags.kill_chain_phases_id.append(SES_KILL_CHAIN_MAPPINGS[kill_chain_phase]) 

        if detection.tags.risk_score < 20:
            detection.tags.risk_level_id = 0
            detection.tags.risk_level = "Info"
        elif detection.tags.risk_score < 40:
            detection.tags.risk_level_id = 1
            detection.tags.risk_level = "Low"            
        elif detection.tags.risk_score < 60:
            detection.tags.risk_level_id = 2
            detection.tags.risk_level = "Medium"   
        elif detection.tags.risk_score < 80:
            detection.tags.risk_level_id = 3
            detection.tags.risk_level = "High"   
        else:
            detection.tags.risk_level_id = 4
            detection.tags.risk_level = "Critical"  

        observable_str = "["
        for i in range(len(detection.tags.observable)):
            role_list = []
            for role in detection.tags.observable[i]["role"]:
                role_list.append(str(SES_OBSERVABLE_ROLE_MAPPING[role]))
            
            observable_str = observable_str + 'create_map("name", "' + detection.tags.observable[i]["name"] + '", "role_ids", [' + ",".join(role_list) + '], "type_id", ' + str(SES_OBSERVABLE_TYPE_MAPPING[detection.tags.observable[i]["type"]]) + ', "value", ' + detection.tags.observable[i]["name"] + ')'
            if not i == len(detection.tags.observable):
                observable_str = observable_str + ', '
        observable_str = observable_str + ']'

        detection.tags.observable_str = observable_str

        j2_env = Environment(
            loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')), 
            trim_blocks=True)
        template = j2_env.get_template('finding_report.j2')
        body = template.render(detection=detection)

        return body



import os

from contentctl.objects.enums import SecurityContentType
from contentctl.output.yml_writer import YmlWriter


class NewContentYmlOutput():
    output_path: str
    
    def __init__(self, output_path:str):
        self.output_path = output_path
    
    
    def writeObjectNewContent(self, object: dict, type: SecurityContentType) -> None:
        if type == SecurityContentType.detections:
            file_path = os.path.join(self.output_path, 'detections', self.convertNameToFileName(object['name'], object['tags']['product']))
            test_obj = {}
            test_obj['name'] = object['name'] + ' Unit Test'
            test_obj['tests'] = [
                {
                    'name': object['name'],
                    'file': self.convertNameToFileName(object['name'],object['tags']['product']),
                    'pass_condition': '| stats count | where count > 0',
                    'earliest_time': '-24h',
                    'latest_time': 'now',
                    'attack_data': [
                        {
                            'file_name': 'UPDATE',
                            'data': 'UPDATE',
                            'source': 'UPDATE',
                            'sourcetype': 'UPDATE',
                            'update_timestamp': True
                        }
                    ]
                }
            ]
            file_path_test = os.path.join(self.output_path, 'tests', self.convertNameToTestFileName(object['name'], object['tags']['product']))
            YmlWriter.writeYmlFile(file_path_test, test_obj)
            #object.pop('source')
            YmlWriter.writeYmlFile(file_path, object)
            print("Successfully created detection " + file_path)
        
        elif type == SecurityContentType.stories:
            file_path = os.path.join(self.output_path, 'stories', self.convertNameToFileName(object['name'], object['tags']['product']))
            YmlWriter.writeYmlFile(file_path, object)
            print("Successfully created story " + file_path)        
        
        else:
            raise(Exception(f"Object Must be Story or Detection, but is not: {object}"))



    def convertNameToFileName(self, name: str, product: list):
        file_name = name \
            .replace(' ', '_') \
            .replace('-','_') \
            .replace('.','_') \
            .replace('/','_') \
            .lower()
        if 'Splunk Behavioral Analytics' in product:
            
            file_name = 'ssa___' + file_name + '.yml'
        else:
            file_name = file_name + '.yml'
        return file_name


    def convertNameToTestFileName(self, name: str, product: list):
        file_name = name \
            .replace(' ', '_') \
            .replace('-','_') \
            .replace('.','_') \
            .replace('/','_') \
            .lower()
        if 'Splunk Behavioral Analytics' in product:          
            file_name = 'ssa___' + file_name + '.test.yml'
        else:
            file_name = file_name + '.test.yml'
        return file_name


import os
import sys
import json
import requests
from requests.auth import HTTPBasicAuth

from dataclasses import dataclass
from configparser import RawConfigParser
import splunklib.client as client

from contentctl.objects.config import Config


@dataclass(frozen=True)
class DeployInputDto:
    path: str
    config: Config


class Deploy:
    def fix_newlines_in_conf_files(self, conf_path: str) -> RawConfigParser:
        parser = RawConfigParser()
        with open(conf_path, "r") as conf_data_file:
            conf_data = conf_data_file.read()

        # ConfigParser cannot read multipleline strings that simply escape the newline character with \
        # To include a newline, you need to include a space at the beginning of the newline.
        # We will simply replace all \NEWLINE with NEWLINESPACE (removing the leading literal \).
        # We will discuss whether we intend to make these changes to the underlying conf files
        # or just apply the changes here
        conf_data = conf_data.replace("\\\n", "\n ")

        parser.read_string(conf_data)
        return parser

    def execute(self, input_dto: DeployInputDto) -> None:

        splunk_args = {
            "host": input_dto.config.deploy.server,
            "port": 8089,
            "username": input_dto.config.deploy.username,
            "password": input_dto.config.deploy.password,
            "owner": "nobody",
            "app": input_dto.config.deploy.app,
        }
        service = client.connect(**splunk_args)

        macros_parser = self.fix_newlines_in_conf_files(
            os.path.join(
                input_dto.path,
                input_dto.config.build.splunk_app.path,
                "default",
                "macros.conf",
            )
        )
        import tqdm

        bar_format_macros = (
            f"Deploying macros        "
            + "{percentage:3.0f}%[{bar:20}]"
            + "[{n_fmt}/{total_fmt} | ETA: {remaining}]"
        )
        bar_format_detections = (
            f"Deploying saved searches"
            + "{percentage:3.0f}%[{bar:20}]"
            + "[{n_fmt}/{total_fmt} | ETA: {remaining}]"
        )
        for section in tqdm.tqdm(
            macros_parser.sections(), bar_format=bar_format_macros
        ):
            try:
                service.post("properties/macros", __stanza=section)
                service.post("properties/macros/" + section, **macros_parser[section])
                # print("Deployed macro: " + section)
            except Exception as e:
                tqdm.tqdm.write(f"Error deploying macro {section}: {str(e)}")

        detection_parser = RawConfigParser()
        detection_parser = self.fix_newlines_in_conf_files(
            os.path.join(
                input_dto.path,
                input_dto.config.build.splunk_app.path,
                "default",
                "savedsearches.conf",
            )
        )
        try:
            service.delete("saved/searches/MSCA - Anomalous usage of 7zip - Rule")
        except Exception as e:
            pass

        for section in tqdm.tqdm(
            detection_parser.sections(), bar_format=bar_format_detections
        ):
            try:
                if section.startswith(input_dto.config.build.splunk_app.prefix):
                    params = detection_parser[section]
                    params["name"] = section
                    response_actions = []
                    if (
                        input_dto.config.detection_configuration.notable
                        and input_dto.config.detection_configuration.notable.rule_description
                    ):
                        response_actions.append("notable")
                    if (
                        input_dto.config.detection_configuration.rba
                        and input_dto.config.detection_configuration.rba.enabled
                    ):
                        response_actions.append("risk")
                    params["actions"] = ",".join(response_actions)
                    params["request.ui_dispatch_app"] = "ES Content Updates"
                    params["request.ui_dispatch_view"] = "ES Content Updates"
                    params["alert_type"] = params.pop("counttype")
                    params["alert_comparator"] = params.pop("relation")
                    params["alert_threshold"] = params.pop("quantity")
                    params.pop("enablesched")

                    service.post("saved/searches", **params)

                    # print("Deployed detection: " + params["name"])
            except Exception as e:
                tqdm.tqdm.write(f"Error deploying saved search {section}: {str(e)}")

        # story_parser = RawConfigParser()
        # story_parser.read(os.path.join(input_dto.path, input_dto.config.build.splunk_app.path, "default", "analyticstories.conf"))

        # for section in story_parser.sections():
        #     if section.startswith("analytic_story"):
        #         params = story_parser[section]
        #         params = dict(params.items())
        #         params["spec_version"] = 1
        #         params["version"] = 1
        #         name = section[17:]
        #         #service.post('services/analyticstories/configs/analytic_story', name=name, content=json.dumps(params))

        #         url = "https://3.72.220.157:8089/services/analyticstories/configs/analytic_story"
        #         data = dict()
        #         data["name"] = name
        #         data["content"] = params
        #         print(json.dumps(data))
        #         response = requests.post(
        #             url,
        #             auth=HTTPBasicAuth('admin', 'fgWFshd0mm7eErMj9qX'),
        #             data=json.dumps(data),
        #             verify=False
        #         )
        #         print(response.text)
